import * as BABYLON from '@babylonjs/core'
import * as gui from '@babylonjs/gui'

export default function calibrate (camera, scene, callback, replay = false) {
  const calibrationTimeStamps = []

  var uiPlane = BABYLON.MeshBuilder.CreatePlane(
    'CalibrationPlane',
    { width: 4, height: 2 },
    scene
  )
  uiPlane.position = new BABYLON.Vector3(0, -0.2, 2)

  uiPlane.parent = camera

  var advancedTexture = gui.AdvancedDynamicTexture.CreateForMesh(
    uiPlane,
    2048,
    1024,
    false
  )
  // ---------------------------
  // step 1: create a calibrate button
  // ---------------------------
  var calibrateButton = gui.Button.CreateSimpleButton('calibrate', 'calibrate')
  calibrateButton.width = '300px'
  calibrateButton.height = '80px'
  calibrateButton.fontSize = '40px'
  calibrateButton.color = 'white'
  calibrateButton.background = 'blue'
  calibrateButton.horizontalAlignment = gui.Control.HORIZONTAL_ALIGNMENT_CENTER
  calibrateButton.verticalAlignment = gui.Control.VERTICAL_ALIGNMENT_CENTER
  advancedTexture.addControl(calibrateButton)

  const startCalibration = function () {
    calibrationTimeStamps.push(Date.now())
    advancedTexture.removeControl(calibrateButton)
    setTimeout(showCrossesSequence, 1000) // delay 1 second to show crosses
  }
  calibrateButton.onPointerUpObservable.add(startCalibration)
  if (replay) {
    startCalibration()
  }

  // ---------------------------
  // Step 2: Create the calibration crosses
  // ---------------------------
  function createCross (i) {
    // trasparent rectangle as a container
    var crossContainer = new gui.Rectangle()
    crossContainer.width = '30px'
    crossContainer.height = '30px'
    crossContainer.background = 'transparent'
    crossContainer.thickness = 0

    // create the crosses
    var horizontalBar = new gui.Rectangle()
    horizontalBar.width = '30px'
    horizontalBar.height = '6px'
    horizontalBar.background = 'white'
    horizontalBar.thickness = 0
    // center alignment
    horizontalBar.horizontalAlignment = gui.Control.HORIZONTAL_ALIGNMENT_CENTER
    horizontalBar.verticalAlignment = gui.Control.VERTICAL_ALIGNMENT_CENTER
    crossContainer.addControl(horizontalBar)

    var verticalBar = new gui.Rectangle()
    verticalBar.width = '6px'
    verticalBar.height = '30px'
    verticalBar.background = 'white'
    verticalBar.thickness = 0
    verticalBar.horizontalAlignment = gui.Control.HORIZONTAL_ALIGNMENT_CENTER
    verticalBar.verticalAlignment = gui.Control.VERTICAL_ALIGNMENT_CENTER
    crossContainer.addControl(verticalBar)

    return crossContainer
  }

  // ---------------------------
  // Step 3: Show calibration crosses in order
  // ---------------------------
  function showCrossesSequence () {
    // definition of positions of crosses: center, topleft, bottomleft, topright, bottomright
    var positions = [
      {
        left: 0,
        top: 0
      },
      {
        left: -424,
        top: -212
      },
      {
        left: -424,
        top: 212
      },
      {
        left: 424,
        top: -212
      },
      {
        left: 424,
        top: 212
      }
    ]

    var index = 0

    function showNextCross () {
      if (index >= positions.length) {
        // after all crosses shown, proceed to the next phase
        showStartButton()
        return
      }

      calibrationTimeStamps.push(Date.now())

      var cross = createCross()
      cross.left = positions[index].left
      cross.top = positions[index].top
      advancedTexture.addControl(cross)

      setTimeout(function () {
        advancedTexture.removeControl(cross)
        index++
        showNextCross()
      }, 2000) // show each cross for 2 seconds
    }

    showNextCross()
  }

  // ---------------------------
  // Step 4: show start button
  // ---------------------------
  function showStartButton () {
    var startButton = gui.Button.CreateSimpleButton('start', 'start')
    startButton.width = '300px'
    startButton.height = '80px'
    startButton.fontSize = '40px'
    startButton.color = 'white'
    startButton.background = 'green'
    startButton.horizontalAlignment = gui.Control.HORIZONTAL_ALIGNMENT_CENTER
    startButton.verticalAlignment = gui.Control.VERTICAL_ALIGNMENT_CENTER
    advancedTexture.addControl(startButton)

    const exitCalibration = function () {
      calibrationTimeStamps.push(Date.now())

      advancedTexture.dispose()
      if (callback && typeof callback === 'function') {
        callback(calibrationTimeStamps)
      }
      uiPlane.dispose()
    }
    startButton.onPointerUpObservable.add(exitCalibration)

    if (replay) {
      setTimeout(exitCalibration, 100)
    }
  }
}
