import questions0 from './questions/questions0.json' with { type: 'json'}
import questions1 from './questions/questions1.json' with { type: 'json'}
import questions2 from './questions/questions2.json' with { type: 'json'}

console.log(
  JSON.stringify([
    analyzeGraph(
      questions0[0].graph,
      questions0[0].highlightedNodes,
      questions0[2].highlightedNodes
    ),
    analyzeGraph(
      questions1[0].graph,
      questions1[0].highlightedNodes,
      questions1[2].highlightedNodes
    ),
    analyzeGraph(
      questions2[0].graph,
      questions2[0].highlightedNodes,
      questions2[2].highlightedNodes
    )
  ])
)

/**
 * 分析图的最短路径、节点连接度和共同邻居
 * @param {Object} graph - 包含nodes和links属性的图对象
 * @param {Array} highlightedNodes1 - 长度为2的数组，表示两个高亮节点的索引，用于最短路径分析
 * @param {Array} highlightedNodes2 - 长度为2的数组，表示两个高亮节点的索引，用于共同邻居分析
 * @returns {Array} 3x2的矩阵，包含各种分析结果
 */
function analyzeGraph (graph, highlightedNodes1, highlightedNodes2) {
  // 创建邻接表表示图结构
  const adjacencyList = {}
  graph.nodes.forEach(node => {
    adjacencyList[node.index] = []
  })

  // 创建节点索引到节点对象的映射，便于快速查找
  const nodeMap = {}
  graph.nodes.forEach(node => {
    nodeMap[node.index] = node
  })

  // 创建边的索引到边对象的映射，便于快速查找
  const linkMap = {}
  graph.links.forEach(link => {
    const sourceIndex = link.source.index
    const targetIndex = link.target.index
    linkMap[`${sourceIndex}-${targetIndex}`] = link
    linkMap[`${targetIndex}-${sourceIndex}`] = link // 对于无向图
  })

  // 添加边到邻接表中
  graph.links.forEach(link => {
    const sourceIndex = link.source.index
    const targetIndex = link.target.index

    adjacencyList[sourceIndex].push(targetIndex)
    adjacencyList[targetIndex].push(sourceIndex)
  })

  // 1. 最短路径分析
  const shortestPathInfo = findShortestPath(
    adjacencyList,
    nodeMap,
    linkMap,
    highlightedNodes1[0],
    highlightedNodes1[1]
  )
  const oneMoreStepPaths = findPathsOneStepLonger(
    adjacencyList,
    nodeMap,
    linkMap,
    highlightedNodes1[0],
    highlightedNodes1[1],
    shortestPathInfo.distance
  )

  // 2. 节点连接度（度）分析
  const nodesWithDegrees = Object.keys(adjacencyList)
    .map(index => ({
      node: nodeMap[index],
      index: parseInt(index),
      degree: adjacencyList[index].length
    }))
    .sort((a, b) => b.degree - a.degree)

  // 找到最大度数
  const maxDegree = nodesWithDegrees[0].degree

  // 找到所有度数最大的节点
  const highestDegreeNodes = nodesWithDegrees
    .filter(item => item.degree === maxDegree)
    .map(item => item.node)

  // 找到第二大度数
  const secondHighestDegree = nodesWithDegrees.find(
    item => item.degree < maxDegree
  )?.degree

  // 找到所有度数并列第二的节点
  const secondHighestDegreeNodes = nodesWithDegrees
    .filter(item => item.degree === secondHighestDegree)
    .map(item => item.node)

  // 获取最高度数节点及其所有链接
  const highestDegreeNodeIndex = highestDegreeNodes[0].index
  const highestDegreeNodeLinks = []

  // 查找与最高度数节点相连的所有边
  graph.links.forEach(link => {
    const sourceIndex = link.source.index
    const targetIndex = link.target.index

    if (
      sourceIndex === highestDegreeNodeIndex ||
      targetIndex === highestDegreeNodeIndex
    ) {
      highestDegreeNodeLinks.push(link)
    }
  })

  // 获取所有第二高度数节点的索引
  const secondHighestDegreeNodeIndices = secondHighestDegreeNodes.map(
    node => node.index
  )
  const secondHighestDegreeNodeLinks = []

  // 查找与第二高度数节点相连的所有边
  graph.links.forEach(link => {
    const sourceIndex = link.source.index
    const targetIndex = link.target.index

    if (
      secondHighestDegreeNodeIndices.includes(sourceIndex) ||
      secondHighestDegreeNodeIndices.includes(targetIndex)
    ) {
      secondHighestDegreeNodeLinks.push(link)
    }
  })

  const highestDegreeInfo = {
    nodes: highestDegreeNodes,
    links: highestDegreeNodeLinks
  }

  const secondHighestDegreeInfo = {
    nodes: secondHighestDegreeNodes,
    links: secondHighestDegreeNodeLinks
  }

  // 3. 共同邻居分析
  const neighbors1 = adjacencyList[highlightedNodes2[0]] || []
  const neighbors2 = adjacencyList[highlightedNodes2[1]] || []

  // 找出共同邻居（交集）
  const commonNeighbors = neighbors1.filter(n => neighbors2.includes(n))

  // 找出所有邻居但排除共同邻居（对称差）
  const allNeighborsExcludingCommon = [
    ...neighbors1.filter(n => !commonNeighbors.includes(n)),
    ...neighbors2.filter(n => !commonNeighbors.includes(n))
  ]

  const commonNeighborsInfo = getNeighborsInfo(
    nodeMap,
    linkMap,
    highlightedNodes2,
    commonNeighbors
  )
  const otherNeighborsInfo = getNeighborsInfo(
    nodeMap,
    linkMap,
    highlightedNodes2,
    allNeighborsExcludingCommon
  )

  // 构建并返回结果矩阵
  return [
    [
      [nodeMap[highlightedNodes1[0]], nodeMap[highlightedNodes1[1]]],
      { nodes: shortestPathInfo.nodes, links: shortestPathInfo.links },
      oneMoreStepPaths
    ], // 最短路径相关
    [
      [],
      highestDegreeInfo,
      secondHighestDegreeInfo
    ], // 节点连接度相关
    [
      [nodeMap[highlightedNodes2[0]], nodeMap[highlightedNodes2[1]]],
      commonNeighborsInfo,
      otherNeighborsInfo
    ] // 共同邻居相关
  ]
}

/**
 * 使用BFS查找两个节点之间的最短路径
 */
function findShortestPath (
  adjacencyList,
  nodeMap,
  linkMap,
  startIndex,
  endIndex
) {
  // 如果起点和终点相同，直接返回该节点
  if (startIndex === endIndex) {
    return {
      nodes: [nodeMap[startIndex]],
      links: [],
      distance: 0
    }
  }

  const visited = new Set()
  const queue = []
  const prev = {}
  const distances = {}

  // 初始化所有节点的距离为无穷大
  Object.keys(adjacencyList).forEach(index => {
    distances[index] = Infinity
  })

  distances[startIndex] = 0
  queue.push(startIndex)
  visited.add(startIndex)

  // BFS搜索
  while (queue.length > 0) {
    const current = queue.shift()

    if (current === endIndex) break

    for (const neighbor of adjacencyList[current]) {
      if (!visited.has(neighbor)) {
        visited.add(neighbor)
        distances[neighbor] = distances[current] + 1
        prev[neighbor] = current
        queue.push(neighbor)
      }
    }
  }

  // 如果没有找到路径
  if (distances[endIndex] === Infinity) {
    return { nodes: [], links: [], distance: Infinity }
  }

  // 重建路径
  const path = []
  let current = endIndex

  while (current !== undefined) {
    path.unshift(current)
    current = prev[current]
  }

  // 获取路径上的节点和边
  const pathNodes = path.map(index => nodeMap[index])
  const pathLinks = []

  for (let i = 0; i < path.length - 1; i++) {
    const link = linkMap[`${path[i]}-${path[i + 1]}`]
    if (link) pathLinks.push(link)
  }

  return {
    nodes: pathNodes,
    links: pathLinks,
    distance: distances[endIndex]
  }
}

/**
 * 查找比最短路径长1的所有路径
 */
function findPathsOneStepLonger (
  adjacencyList,
  nodeMap,
  linkMap,
  startIndex,
  endIndex,
  shortestDistance
) {
  if (shortestDistance === Infinity) {
    return { nodes: [], links: [] }
  }

  const targetDistance = shortestDistance + 1
  const nodeIndices = new Set()
  const linkKeys = new Set()

  function dfs (node, path, distance) {
    // 如果达到目标距离
    if (distance === targetDistance) {
      // 检查是否到达目标节点
      if (node === endIndex) {
        // 记录此路径
        for (const nodeIndex of path) {
          nodeIndices.add(nodeIndex)
        }

        for (let i = 0; i < path.length - 1; i++) {
          const linkKey = `${path[i]}-${path[i + 1]}`
          if (linkMap[linkKey]) linkKeys.add(linkKey)
        }
      }
      return
    }

    // 如果还没有达到目标距离，继续探索
    for (const neighbor of adjacencyList[node]) {
      // 避免路径中出现环
      if (!path.includes(neighbor)) {
        path.push(neighbor)
        dfs(neighbor, path, distance + 1)
        path.pop() // 回溯
      }
    }
  }

  // 从起始节点开始DFS
  dfs(startIndex, [startIndex], 0)

  // 转换为实际对象
  const resultNodes = Array.from(nodeIndices).map(index => nodeMap[index])
  const resultLinks = Array.from(linkKeys).map(key => linkMap[key])

  return {
    nodes: resultNodes,
    links: resultLinks
  }
}

/**
 * 获取邻居节点和相关边的信息
 */
function getNeighborsInfo (nodeMap, linkMap, highlightedNodes, neighbors) {
  const nodeIndices = new Set()
  const linkKeys = new Set()

  // 添加高亮节点
  nodeIndices.add(highlightedNodes[0])
  nodeIndices.add(highlightedNodes[1])

  // 添加邻居节点和边
  neighbors.forEach(neighborIndex => {
    nodeIndices.add(neighborIndex)

    // 查找高亮节点与此邻居之间的边
    const link1Key = `${highlightedNodes[0]}-${neighborIndex}`
    const link2Key = `${highlightedNodes[1]}-${neighborIndex}`

    if (linkMap[link1Key]) linkKeys.add(link1Key)
    if (linkMap[link2Key]) linkKeys.add(link2Key)
  })

  nodeIndices.delete(highlightedNodes[0])
  nodeIndices.delete(highlightedNodes[1])
  // 转换为实际对象
  const resultNodes = Array.from(nodeIndices).map(index => nodeMap[index])
  const resultLinks = Array.from(linkKeys).map(key => linkMap[key])

  return {
    nodes: resultNodes,
    links: resultLinks
  }
}
