import * as anu from '@jpmorganchase/anu'
import * as BABYLON from '@babylonjs/core'

export default function draw (question, scene) {
  const data = question.graph
  console.log(data)

  const highlightedNodes = new Set(question.highlightedNodes)

  //We use Mesh instancing here for better performance, first we create a Mesh that serves as the root Node
  let rootSphere = anu.create('sphere', 'node', { diameter: 0.01 })
  rootSphere.isVisible = false
  rootSphere.registerInstancedBuffer('color', 4) //We need an InstancedBuffer to set the color of instances

  //Create a Center of Transform TransformNode that serves the parent node for all our meshes that make up our chart
  let CoT = anu.create('cot', 'cot')

  rootSphere.parent = CoT

  //Select our CoT so that we have it as a Selection object
  let network = anu.selectName('cot', scene)

  //Create instanced sphere meshes from our rootSphere as children of our CoT for each row of our data and set their initial visual encodings using method chaining
  let nodes = network
    .bindInstance(rootSphere, data.nodes)
    .position(d => new BABYLON.Vector3(d.x, d.y, d.z))
    .setInstancedBuffer('color', d =>
      highlightedNodes.has(d.index)
        ? new BABYLON.Color4(1, 0, 0, 1)
        : new BABYLON.Color4(1, 1, 0, 1)
    )
    .id(d => d.index)

  //Create a helper function that will return us an array of arrays where each sub-array is the start and end Vector3 of each link
  function dataToLinks (data) {
    let lines = []
    data.forEach((v, i) => {
      let start = new BABYLON.Vector3(v.source.x, v.source.y, v.source.z)
      let end = new BABYLON.Vector3(v.target.x, v.target.y, v.target.z)
      lines.push([start, end])
    })
    return lines
  }

  //Create our lineSystem mesh using our data and helper function from above
  let links = network
    .bind('lineSystem', { lines: d => dataToLinks(d), updatable: true }, [
      data.links
    ])
    .prop('color', new BABYLON.Color4(1, 1, 1, 1))

  let labels = createLabelsWithSPS(scene, data.nodes, highlightedNodes)
  labels.parent = CoT

  return CoT
}

function createLabelsWithSPS (scene, nodes, highlightedNodes) {
  // 设置一些常量
  const labelSizePx = 128 // 每个标签的大小(像素)
  const fontSizePx = 40 // 字体大小(像素)
  const font = `bold ${fontSizePx}px monospace`
  const boxSize = 0.05 // 标签平面的大小(3D空间单位)

  // 获取引擎的最大纹理尺寸
  const maxTextureSizePx = 1280
  const maxLabelsPerDimension = Math.floor(maxTextureSizePx / labelSizePx)
  const maxLabelsPerTexture = Math.pow(maxLabelsPerDimension, 2)

  // 创建源平面 - 这将作为SPS的基础
  const sourcePlane = BABYLON.MeshBuilder.CreatePlane(
    'labelPlane',
    {
      width: boxSize,
      height: boxSize
    },
    scene
  )
  sourcePlane.isVisible = false

  // 设置SPS系统
  const SPS = new BABYLON.SolidParticleSystem('labelSPS', scene, {
    updatable: false,
    enableMultiMaterial: true
  })

  // 跟踪材质和纹理
  const materials = []
  let texture
  let textureIndex = -1
  let nodeIndexInsideTexture = 0
  let labelPos

  // 添加所有节点到SPS
  for (let i = 0; i < nodes.length; i++) {
    // 当达到每个纹理的最大标签数时，创建一个新纹理
    if (i % maxLabelsPerTexture === 0) {
      // 如果已经有一个纹理，更新它
      if (texture) {
        texture.update()
      }

      // 创建新纹理
      textureIndex++
      texture = new BABYLON.DynamicTexture(
        'label-texture' + textureIndex,
        { width: maxTextureSizePx, height: maxTextureSizePx },
        scene,
        false
      )

      // 为新纹理创建材质
      const material = new BABYLON.StandardMaterial(
        'label-material' + textureIndex,
        scene
      )
      material.diffuseTexture = texture
      material.diffuseTexture.hasAlpha = true
      material.specularColor = new BABYLON.Color3(0, 0, 0)
      material.emissiveColor = new BABYLON.Color3(1, 1, 1)
      material.backFaceCulling = false

      materials.push(material)
      nodeIndexInsideTexture = 0
    }

    // 绘制此节点的标签文本
    const node = nodes[i]
    labelPos = drawLabelText(
      texture,
      node.index.toString(),
      nodeIndexInsideTexture,
      maxLabelsPerDimension,
      labelSizePx,
      fontSizePx,
      font,
      highlightedNodes
    )

    // 定义SPS粒子的位置和UV函数
    const positionFunction = function (particle, particleIndex, particleShape) {
      // 设置位置
      particle.position.x = node.x - 0.02
      particle.position.y = node.y + 0.01
      particle.position.z = node.z
      // particle.rotation.y = Math.PI / 2

      // 设置UV坐标，从纹理中获取正确的标签部分
      particle.uvs.x = labelPos.left / maxTextureSizePx
      particle.uvs.y = 1 - labelPos.bottom / maxTextureSizePx
      particle.uvs.z = labelPos.right / maxTextureSizePx
      particle.uvs.w = 1 - labelPos.top / maxTextureSizePx

      // 设置材质索引
      particle.materialIndex = textureIndex
    }

    // 添加平面形状到SPS
    SPS.addShape(sourcePlane, 1, { positionFunction: positionFunction })

    nodeIndexInsideTexture++
  }

  // 更新最后一个纹理
  if (texture) {
    texture.update()
  }

  sourcePlane.dispose()

  // 构建SPS网格
  const mesh = SPS.buildMesh()
  // mesh.billboardMode = BABYLON.Mesh.BILLBOARDMODE_ALL // 使所有标签面向相机
  SPS.setMultiMaterial(materials)

  return mesh
}

// 辅助函数：在纹理上绘制标签文本
function drawLabelText (
  texture,
  label,
  index,
  maxLabelsPerDimension,
  labelSizePx,
  fontSizePx,
  font,
  highlightedNodes
) {
  // 计算标签在纹理中的行和列
  const row = Math.floor(index / maxLabelsPerDimension)
  const col = index % maxLabelsPerDimension

  // 计算标签在纹理中的位置
  const left = labelSizePx * col
  const right = left + labelSizePx
  const top = labelSizePx * row
  const bottom = top + labelSizePx

  // 获取纹理的上下文并绘制背景（可选）
  const textureContext = texture.getContext()

  // 清除这个标签区域（透明背景）
  textureContext.clearRect(left, top, labelSizePx, labelSizePx)

  // 绘制文本
  texture.drawText(
    label,
    left + labelSizePx / 2, // 文本中心X
    top + labelSizePx / 2, // 文本中心Y
    font,
    highlightedNodes.has(+label) ? 'red' : 'yellow',
    'transparent',
    true, // 居中对齐
    true // 垂直居中
  )

  return { left, right, top, bottom }
}
