import {
  Scene,
  HemisphericLight,
  PointLight,
  UniversalCamera,
  VideoRecorder,
  Vector3,
  Color3,
  Color4,
  MeshBuilder
} from '@babylonjs/core'
import * as gui from '@babylonjs/gui'

//Create and export a function that takes a Babylon engine and returns a Babylon Scene
export async function createScene (engine, fixations, fileIndex) {
  //Create an empty Scene
  const scene = new Scene(engine)
  scene.clearColor = new Color4(0, 0, 0, 0.8)

  const pointLight = new PointLight('light', new Vector3(0, 3.4, 0), scene)
  pointLight.intensity = 0.8
  //Add some lighting
  let light = new HemisphericLight('light1', new Vector3(0, 10, 0), scene)
  light.diffuse = new Color3(1, 1, 1)
  light.specular = new Color3(1, 1, 1)
  light.groundColor = new Color3(1, 1, 1)

  //Add a camera and adjust its properties afterwards
  const camera = new UniversalCamera('camera', new Vector3(0, 0, 0), scene)
  camera.minZ = 0 // Adjust the distance of the camera's near plane
  // camera.position = new Vector3(-1.2, 0.2, -1)
  camera.fov = 1.45
  camera.attachControl(true) // Allow the camera to respond to user controls
  camera.inputs.attached.mouse.detachControl() // Disable mouse control

  var uiPlane = MeshBuilder.CreatePlane(
    'uiPlane',
    { width: 8, height: 4 },
    scene
  )
  uiPlane.position = new Vector3(0, 0, 2)

  uiPlane.renderingGroupId = 1

  uiPlane.parent = camera

  var advancedTexture = gui.AdvancedDynamicTexture.CreateForMesh(
    uiPlane,
    4096,
    2048,
    false
  )

  const firstFrameTimestamp = fixations[0].timestamp
  console.log(firstFrameTimestamp)

  // 计算最后一个fixation的时间戳和持续时间，用于安排录制结束
  const lastFixation = fixations[fixations.length - 1]
  const lastFixationTime = lastFixation.timestamp - firstFrameTimestamp
  const lastFixationDuration = lastFixation.duration

  // 用于存储活跃的fixation点的数组，包含圆圈、标签、位置信息和连线
  const activeFixations = []
  // 用于存储scanpath线条的数组
  const scanpathLines = []

  let recorder
  if (VideoRecorder.IsSupported(engine)) {
    recorder = new VideoRecorder(engine)
    recorder.startRecording(`fixations_recording_${fileIndex}.webm`, 3600)

    setTimeout(() => {
      console.log('Stopping recording after last fixation')
      recorder.stopRecording()
    }, lastFixationTime + Math.max(lastFixationDuration, 2000) + 10000) // 最后fixation时间 + 持续时间 + 10秒
  }

  for (let i = 0; i < fixations.length; i++) {
    const { left, top, duration, timestamp, index } = fixations[i]

    setTimeout(() => {
      drawFixationWithPath(left, top, duration, index)
    }, timestamp - firstFrameTimestamp)
  }

  function drawFixationWithPath (left, top, duration, index) {
    // 创建圆圈
    let circle = new gui.Ellipse()
    circle.width = '50px'
    circle.height = '50px'
    circle.color = 'blue'
    circle.thickness = 3
    circle.background = 'transparent'
    circle.left = left
    circle.top = top
    advancedTexture.addControl(circle)

    // 创建标签
    let label = new gui.TextBlock()
    label.text = index.toString()
    label.color = 'white'
    label.fontSize = '25px'
    label.left = left + 30
    label.top = top + 30
    advancedTexture.addControl(label)

    // 创建新的fixation对象
    const newFixation = {
      circle,
      label,
      left,
      top,
      index,
      lines: [] // 存储与该fixation相连的线
    }

    // 如果有前一个fixation点，则画一条连线
    if (activeFixations.length > 0) {
      const prevFixation = activeFixations[activeFixations.length - 1]

      // 创建连线
      let line = new gui.Line()
      line.x1 = prevFixation.left + 2048
      line.y1 = prevFixation.top + 1024
      line.x2 = left + 2048
      line.y2 = top + 1024
      line.color = 'white'
      line.lineWidth = 5
      advancedTexture.addControl(line)

      // 将线添加到两个fixation点的lines数组中
      prevFixation.lines.push(line)
      newFixation.lines.push(line)

      // 将线添加到scanpathLines数组中
      scanpathLines.push(line)
    }

    // 将新的fixation点添加到活跃数组中
    activeFixations.push(newFixation)

    // 延长移除时间，使用Math.max(duration, 5000)
    setTimeout(function () {
      // 从advancedTexture中移除圆圈和标签
      advancedTexture.removeControl(circle)
      advancedTexture.removeControl(label)

      // 移除与该fixation相连的所有线
      for (const line of newFixation.lines) {
        advancedTexture.removeControl(line)
        // 从scanpathLines数组中移除该线
        const lineIndex = scanpathLines.indexOf(line)
        if (lineIndex !== -1) {
          scanpathLines.splice(lineIndex, 1)
        }
      }

      // 从activeFixations数组中移除该fixation
      const fixationIndex = activeFixations.indexOf(newFixation)
      if (fixationIndex !== -1) {
        activeFixations.splice(fixationIndex, 1)
      }
    }, Math.max(duration, 2000))
  }

  return scene
}
