import trimesh
import numpy as np
from trimesh.ray.ray_triangle import RayMeshIntersector
import pyglet
from pyglet import gl
# from trimesh.ray import ray_pyembree  # only works on Linux and Windows

def plot_meshes(meshes: list, ray_cones: list):
    """
    Plot the meshes and the gaze ray using pyglet.
    """
    scene = trimesh.Scene()
    for _, mesh in enumerate(meshes):
        # Add each mesh to the scene with a unique name
        scene.add_geometry(mesh)

    for id, ray_cone in enumerate(ray_cones):
        # Plot the ray
        ray_origin, ray_direction = ray_cone
        length = 2.0  # Adjust the ray length
        end = ray_origin + ray_direction / np.linalg.norm(ray_direction) * length
        ray_path = trimesh.load_path(np.array([ray_origin, end]))
        origin_sphere = trimesh.primitives.Sphere(radius=0.1, center=ray_origin)
        scene.add_geometry(origin_sphere, node_name="origin_marker")
        scene.add_geometry(ray_path, node_name=f"ray_{id}")

    viewer = scene.show(
    smooth=True,           # Enable smooth shading
    flags={
        'axis': False,     # Hide axis for cleaner look
        'grid': False,     # Hide grid
        'wireframe': False # Set to True if you want wireframe mode
    },
    resolution=(1920, 1080),  # Higher resolution
    window_conf=pyglet.gl.Config(
        sample_buffers=1,      # Enable multisampling
        samples=16,            # 4x anti-aliasing
        depth_size=24,         # 24-bit depth buffer
        double_buffer=True     # Double buffering
    )
)
    
    return viewer

def create_question_plane():
    """
    Create a plane-like mesh for the question.

    Returns:
        trimesh object representing the question plane.
    """
    # Create a plane-like mesh: thin box (X=width, Y=thickness, Z=height)
    question_plane = trimesh.creation.box(extents=(0.5, 1.0, 1e-3))
    # RH equivalent of LH rotation: +π (180°) around Y-axis
    # rotation_matrix = trimesh.transformations.rotation_matrix(
    #     angle=-np.pi/2,  # +90° for RH
    #     direction=[1, 0, 0],  # Z-axis
    #     point=[0, 0, 0]
    # )
    # RH equivalent of position: flip Z from -1.5 to +1.5
    translation_matrix = trimesh.transformations.translation_matrix([0, 0.15, -1.5])
    # Apply transformation: rotate, then translate
    # transform = translation_matrix @ rotation_matrix
    transform = translation_matrix
    question_plane.apply_transform(transform)
    question_plane.metadata['index'] = 'question_plane'
    return [question_plane]

def create_node_meshes(nodes, radius=0.025, color_map=None):
    """
    Create nodes as spheres.

    Args:
        nodes: List of node dictionaries with 'x', 'y', 'z' coordinates. Note that input is left-handed.
        radius: Radius of the sphere.
        color_map: Dictionary mapping metadata index to color (key = node index, value = color).
                  Color can be RGB tuple (r, g, b) with values 0-255 or RGBA tuple (r, g, b, a).

    Returns:
       list of Trimesh spheres
    """
    meshes = []
    for node in nodes:
        sphere = trimesh.creation.icosphere(subdivisions=2, radius=radius)
        # flip in Z for right-handed system in Python
        sphere.apply_translation([node["x"], node["y"], node["z"]])
        sphere.metadata['index'] = node['index']
        
        # Apply color if color_map is provided
        if color_map and node['index'] in color_map:
            color = color_map[node['index']]
            # Ensure color is in RGBA format
            if len(color) == 3:
                color = (*color, 255)  # Add alpha channel
            sphere.visual.face_colors = color
        
        meshes.append(sphere)
    return meshes

def create_edge_meshes(edges, radius=0.002, color_map=None):
    """
    Create edges as cylinders.

    Args:
        edges: List of edge dictionaries with source and target nodes containing 'x', 'y', 'z' coordinates.
        radius: Radius of the cylinder.
        color_map: Dictionary mapping metadata index to color (key = "{source_index}-{target_index}", value = color).
                  Color can be RGB tuple (r, g, b) with values 0-255 or RGBA tuple (r, g, b, a).

    Returns:
       list of Trimesh cylinders
    """
    meshes = []
    for edge in edges:
        p1 = np.array([edge["source"]["x"], edge["source"]["y"], edge["source"]["z"]])
        p2 = np.array([edge["target"]["x"], edge["target"]["y"], edge["target"]["z"]])
        if np.array_equal(p1, p2):
            continue
        cylinder = create_cylinder_between(p1, p2, radius=radius)
        edge_index = f"{edge['source']['index']}-{edge['target']['index']}"
        cylinder.metadata['index'] = edge_index
        
        # Apply color if color_map is provided
        if color_map and edge_index in color_map:
            color = color_map[edge_index]
            # Ensure color is in RGBA format
            if len(color) == 3:
                color = (*color, 255)  # Add alpha channel
            cylinder.visual.face_colors = color
        
        meshes.append(cylinder)
    return meshes

def create_cylinder_between(p1, p2, radius, sections=32):
    """
    Create a cylinder mesh between two 3D points.

    Args:
        p1, p2: Endpoints of the cylinder (3D coordinates).
        radius: Radius of the cylinder.
        sections: Number of radial segments.

    Returns:
        Trimesh cylinder aligned between p1 and p2.
    """
    # Vector from p1 to p2
    vec = np.array(p2) - np.array(p1)
    height = np.linalg.norm(vec)

    if height < 1e-8:
        raise ValueError("Points are too close together to form a valid cylinder.")

    # Create cylinder aligned along +Z
    cylinder = trimesh.creation.cylinder(radius=radius, height=height, sections=sections)

    # Compute transform to align +Z to the vector
    z_axis = np.array([0, 0, 1])
    vec_norm = vec / height

    rotation = trimesh.geometry.align_vectors(z_axis, vec_norm)
    cylinder.apply_transform(rotation)

    # Move to midpoint between p1 and p2
    midpoint = (np.array(p1) + np.array(p2)) / 2
    cylinder.apply_translation(midpoint)

    return cylinder

def multi_pick_with_ray(meshes, ray_origin, ray_direction):
    """
    Ray-pick multiple meshes and return all hit objects with details.

    Args:
        meshes: list of trimesh.Trimesh objects
        ray_origin: (3,) ray origin
        ray_direction: (3,) normalized ray direction

    Returns:
        List of hit results with (mesh_index, location, distance)
    """
    hits = []
    for i, mesh in enumerate(meshes):
        # intersector = ray_pyembree.RayMeshIntersector(mesh)
        intersector = RayMeshIntersector(mesh)

        # Run intersection
        locations, index_ray, index_tri = intersector.intersects_location(
            [ray_origin], [ray_direction]
        )

        # Store all hits (can be multiple if ray enters and exits mesh)
        for loc in locations:
            distance = np.linalg.norm(loc - ray_origin)
            hits.append((mesh.metadata['index'], loc, distance))

    # Sort hits by distance
    hits.sort(key=lambda x: x[2])
    return hits

def calculate_ray(camera_location, camera_rotation, azimuth=0, elevation=0):
    """
    Calculate ray direction from camera location and rotation, with additional azimuth and elevation angles.

    Args:
        camera_location: (3,) camera position
        camera_rotation: (4,) quaternion rotation
        azimuth: horizontal angle in degrees (rotation around vertical axis)
        elevation: vertical angle in degrees (up/down rotation)

    Returns:
        ray_origin: (3,) origin of the ray
        ray_direction: (3,) direction of the ray
    """
    # Convert quaternion to rotation matrix
    q = np.array(camera_rotation)
    x, y, z, w = q
    rotation_new = trimesh.transformations.quaternion_matrix([w, x, y, z])

    # rotation_matrix = np.array([
    #     [1 - 2*(y**2 + z**2), 2*(x*y - w*z), 2*(x*z + w*y)],
    #     [2*(x*y + w*z), 1 - 2*(x**2 + z**2), 2*(y*z - w*x)],
    #     [2*(x*z - w*y), 2*(y*z + w*x), 1 - 2*(x**2 + y**2)]
    # ])

    # Convert azimuth and elevation from degrees to radians
    azimuth_rad = np.radians(azimuth)
    elevation_rad = -np.radians(elevation)

    # In camera space, local forward is [0, 0, -1] for left-handed system
    local_forward = np.array([0, 0, 1])

    # Calculate the local direction with azimuth and elevation
    # Apply elevation first (pitch, rotation around X-axis)
    cos_elev = np.cos(elevation_rad)
    sin_elev = np.sin(elevation_rad)
    elevation_matrix = np.array([
        [1, 0, 0],
        [0, cos_elev, -sin_elev],
        [0, sin_elev, cos_elev]
    ])

    # Then apply azimuth (yaw, rotation around Y-axis)
    cos_azim = np.cos(azimuth_rad)
    sin_azim = np.sin(azimuth_rad)
    azimuth_matrix = np.array([
        [cos_azim, 0, sin_azim],
        [0, 1, 0],
        [-sin_azim, 0, cos_azim]
    ])

    # Apply rotations to local forward: first elevation, then azimuth
    local_direction = azimuth_matrix @ elevation_matrix @ local_forward

    # Transform to world coordinates using the camera's rotation matrix
    ray_direction = rotation_new[:3,:3] @ local_direction

    # Normalize the ray direction
    ray_direction /= np.linalg.norm(ray_direction)

    ray_origin = np.array([camera_location[0], camera_location[1], camera_location[2]])

    return ray_origin, ray_direction


def create_cone_rays(origin, direction, cone_angle, num_rays):
    """
    Create a cone of rays surrounding the original ray
    
    Parameters:
        origin: Ray origin point
        direction: Ray direction vector
        cone_angle: Cone angle in radians
        num_rays: Number of additional rays to create
    
    Returns:
        List of rays as (origin, direction) tuples
    """
    # Normalize direction vector
    direction = direction / np.linalg.norm(direction)
    
    # Find two vectors perpendicular to the direction
    if abs(direction[0]) < abs(direction[1]):
        perpendicular = np.array([1, 0, 0])
    else:
        perpendicular = np.array([0, 1, 0])
    
    # Create orthogonal vectors
    perpendicular = np.cross(direction, perpendicular)
    perpendicular = perpendicular / np.linalg.norm(perpendicular)
    perpendicular2 = np.cross(direction, perpendicular)
    
    # Create ray list
    rays = [(origin, direction)]  # Add the original ray first
    
    # Create additional rays in a cone distribution
    for i in range(num_rays):
        # Calculate angle for this ray
        theta = 2 * np.pi * i / num_rays
        
        # Calculate offset direction for this ray
        offset_direction = (
            direction * np.cos(cone_angle) +
            np.sin(cone_angle) * (
                perpendicular * np.cos(theta) + 
                perpendicular2 * np.sin(theta)
            )
        )
        
        # Normalize the offset direction
        offset_direction = offset_direction / np.linalg.norm(offset_direction)
        
        # Add this ray to the list
        rays.append((origin, offset_direction))
    
    return rays

def multi_pick_with_cone(all_meshes, ray_origin, ray_direction, cone_angle=0.05, num_rays=8, remove_duplicates=True):
    """
    Perform ray picking using a cone of rays with duplicate removal
    
    Parameters:
        all_meshes: Meshes to check for intersections
        ray_origin: Original ray origin point
        ray_direction: Original ray direction vector
        cone_angle: Cone angle in radians (default: 0.05 radians, ~2.9 degrees)
        num_rays: Number of additional rays to create (default: 8)
        remove_duplicates: Whether to remove duplicate hits (default: True)
    
    Returns:
        List of hits from all rays, with optional duplicate removal
    """
    # Create cone rays
    cone_rays = create_cone_rays(ray_origin, ray_direction, cone_angle, num_rays)
    
    # Collect all hits
    all_hits = []
    
    # Perform ray picking for each ray
    for origin, direction in cone_rays:
        hits = multi_pick_with_ray(all_meshes, origin, direction)
        all_hits.extend(hits)
    
    # Return all hits if not removing duplicates
    if not remove_duplicates:
        return all_hits
    
    # Process to remove duplicate hits - keep only closest hit per mesh ID
    hits_by_mesh = {}
    for hit in all_hits:
        mesh_id, location, distance = hit
        
        # Convert mesh_id to a hashable type if needed
        mesh_id_key = str(mesh_id) if isinstance(mesh_id, (list, dict, set)) else mesh_id
        
        # If this mesh_id is not in hits_by_mesh, or this hit is closer than existing one
        if mesh_id_key not in hits_by_mesh or distance < hits_by_mesh[mesh_id_key][2]:
            hits_by_mesh[mesh_id_key] = (mesh_id, location, distance)
    
    # Convert back to list
    return list(hits_by_mesh.values()), cone_rays