import argparse
import json
import pandas as pd
import numpy as np
import os
from graph import create_node_meshes, create_edge_meshes, plot_meshes


def read_network_gaze_data(dataset_dir='/netpool/homes/wangyo/Dataset/NetworkGaze', G=2, T=2, p=1):
    """
    Read NetworkGaze dataset CSV file based on parameters G, T, and p.

    Parameters:
    -----------
    dataset_dir : str, optional
        Base directory of the dataset, default is '/netpool/homes/wangyo/Dataset/NetworkGaze'
    G : int
        Graph ID
    T : int
        Task ID
    p : int
        participant ID

    Returns:
    --------
    pandas.DataFrame
        DataFrame containing the data from the CSV file
    """
    file_path = os.path.join(dataset_dir, 'Processed',
                             f'G{str(G)}', f'T{str(T)}', f'p{p}.csv')

    if not os.path.exists(file_path):
        raise FileNotFoundError(f"File not found at path: {file_path}")

    df = pd.read_csv(file_path)

    print(f"Successfully loaded data with shape: {df.shape}")

    return df


def process_hits(hits_str):
    """Process hits string, returns list of (type, distance) tuples"""
    if pd.isna(hits_str) or hits_str == '':
        return []

    hits = []
    for hit in hits_str.split(';'):
        if hit and '|' in hit:  # Ensure hit is not empty and contains separator
            try:
                id_str, len_str = hit.split('|')
                hit_id = id_str if "-" in id_str else int(id_str)
                weight = 1 / float(len_str)
                hits.append((hit_id, weight))
            except (ValueError, IndexError):
                # Skip malformed hits
                continue

    return hits


def create_color_map_gradient(value_map, outlier_threshold=1):
    """
    Handle outliers by capping extreme values at a percentile threshold.

    Args:
        value_map: Dictionary with keys of any type and non-negative float values
        outlier_threshold: Percentile threshold (0.9 = 90th percentile, 0.95 = 95th percentile)

    Returns:
        color_map: Dictionary with same keys as input, RGB color tuples (r, g, b)
    """
    if not value_map:
        return {}

    values = list(value_map.values())
    max_value = max(values)
    min_value = min(values)

    if max_value == min_value:
        return {key: (255, 255, 255) for key in value_map.keys()}

    # Use percentile as the effective maximum
    effective_max = np.percentile(values, outlier_threshold * 100)

    color_map = {}
    white = np.array([255, 255, 255])
    dark_blue = np.array([0, 0, 139])

    for key, value in value_map.items():
        # Cap the value at effective_max
        capped_value = min(value, effective_max)
        normalized_value = (capped_value - min_value) / \
            (effective_max - min_value)

        color = white + normalized_value * (dark_blue - white)
        color = np.clip(color, 0, 255).astype(int)
        color_map[key] = tuple(color)

    return color_map


def heat_map(mesh_id_sets, graphs, graph_id, task_id):
    mesh_id_set = mesh_id_sets[graph_id]
    mesh_map = {key: 0 for key in mesh_id_set}

    for i in range(1, 19):
        df = read_network_gaze_data(
            dataset_dir=args.dataset_dir,
            G=graph_id,
            T=task_id,
            p=i
        )
        df = df.copy()
        # Process hits for all rows
        df['processed_hits'] = df['hits2'].apply(process_hits)
        df = df[df['processed_hits'].apply(len) > 0]
        for i, row in enumerate(df.iterrows()):
            idx, row_data = row
            hits = row_data['processed_hits']
            if hits:
                for hit_id, weight in hits:
                    mesh_map[hit_id] += weight

    color_map = create_color_map_gradient(mesh_map)

    graph = graphs[graph_id]
    all_meshes = create_node_meshes(graph["nodes"], radius=0.005, color_map=color_map) + \
        create_edge_meshes(graph["edges"], radius=0.0005, color_map=color_map)
    plot_meshes(all_meshes, ())


def main(args):
    mesh_id_sets = []
    graphs = []
    for i in range(3):
        with open(f'../data/questions/questions{i}.json', encoding='utf-8') as json_data:
            graph_json = json.load(json_data)
            nodes = graph_json[0]['graph']['nodes']
            edges = graph_json[0]['graph']['links']
            graphs.append({"nodes": nodes, "edges": edges})

            node_id = [node['index']
                       for node in nodes]
            edge_id = [
                f"{edge['source']['index']}-{edge['target']['index']}" for edge in edges]
            mesh_id_set = set([*node_id, *edge_id])
            mesh_id_sets.append(mesh_id_set)

    for i in range(3):
        with open(f'../data/questions/questions{i}.json', encoding='utf-8') as json_data:
            graph_json = json.load(json_data)

    heat_map(mesh_id_sets, graphs, int(args.G), int(args.T))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Read NetworkGaze dataset CSV files')

    parser.add_argument('--dataset_dir', type=str,
                        default='/netpool/homes/wangyo/Dataset/NetworkGaze',
                        help='Base directory of the dataset')
    parser.add_argument('--G', type=str, default=2,
                        help='Graph ID')
    parser.add_argument('--T', type=str, default=2,
                        help='Task ID')
    parser.add_argument('--p', type=str, default=1,
                        help='participant ID')

    args = parser.parse_args()
    main(args)
