import os
import math
import argparse
import numpy as np
import pandas as pd
import json
from graph import create_node_meshes, create_edge_meshes, create_question_plane, calculate_ray, plot_meshes, multi_pick_with_cone
from gaze_utils import calc_saccade_len_vel
from utils import get_tasks, update_fix_cnt, find_camera_pose, statistical_tests

def process_scene(graph_id: int, task_id: int, questionnaire: dict, df_task: pd.DataFrame, t_offset: np.int64, plot_mesh: bool, all_meshes, related_meshes):
    """ process the scene in every camera location
    Args:
        questionnaire (json): questionnaire and camera poses
        df_task (pd.DataFrame): gaze data
        t_offset (int): offset between gaze data and camera data
    Return:
        fix_cnt (list): [fix on at least a mesh, fix on highlight meshes, fix on task meshes, fix on confusion meshes]
        first_time_on_task (float): first time on task, seconds
    """
    print(f'Processing scene {graph_id}, task {task_id}...')

    # first time on task
    ft_smp = {"start": None, "flag": False, "interval":0}
    # [fix on any graph mesh, fix on highlight meshes, fix on task meshes, fix on confusion meshes, fix on question plane]
    fix_cnt = [0,0,0,0,0,0]
    hits_stat_arr = []
    hits_stat_arr2 = []

    for idx, task in df_task.iterrows():
        # if idx < 633: continue
        camera_loc = find_camera_pose(questionnaire, task['start timestamp [ns]'] + t_offset, task['end timestamp [ns]'] + t_offset)
        if camera_loc is None: # cannot find camera pose
            hits_stat_arr.append("")
            hits_stat_arr2.append("")
            continue

        # Set cone parameters
        cone_angle = 0.05  # 0.05 radians is about 2.9 degrees - adjust as needed
        num_rays = 8       # Number of additional rays - adjust as needed

        ray_origin, ray_direction = calculate_ray(
            camera_loc['position'], camera_loc['rotation'], task['azimuth [deg]'], task['elevation [deg]'])

        # Perform ray picking with the cone of rays and remove duplicates
        hits, cone_rays = multi_pick_with_cone(all_meshes, ray_origin, ray_direction, cone_angle, num_rays, remove_duplicates=True)

        if plot_mesh:
            plot_meshes(all_meshes, cone_rays)

        if len(hits) == 0:
            hits_stat_arr.append("")
            hits_stat_arr2.append("")
            continue

        fix_flag = [False, False, False, False, False]
        hits_stat = []
        hits_stat2 = []

        for _, (hit_mesh_id, location, distance) in enumerate(hits):
            # Process each hit as in the original code
            if hit_mesh_id == 'question_plane':
                fix_flag[4] = True
                hits_stat.append({"type": 0, "distance": distance})
            else:
                hits_stat2.append({"id":hit_mesh_id, "len":len(hits)})
                if not ft_smp['start']:
                    ft_smp['start'] = np.int64(camera_loc['timeStamp'])
                update_fix_cnt(ft_smp, fix_flag, camera_loc, hit_mesh_id, related_meshes, task_id, hits_stat, distance)

        for i in range(len(fix_flag)) :
            fix_cnt[i] += int(fix_flag[i])
        fix_cnt[5] += int(fix_flag[1] | fix_flag[2] | fix_flag[3])
        hits_stat_arr.append(';'.join([f"{mesh['type']}|{mesh['distance']}" for mesh in hits_stat]))
        hits_stat_arr2.append(';'.join([f"{mesh['id']}|{mesh['len']}" for mesh in hits_stat2]))

    df_task["hits"] = hits_stat_arr
    df_task["hits2"] = hits_stat_arr2
    return fix_cnt, ft_smp

def main(dataset_dir: str, save_fixations: bool, plot_mesh: bool):
    """
    Args:
        dataset_dir (str): dataset directory containing participant's gaze data
        save_fixations (bool): if True, save fixations into separate tasks
    """

    graphs = []
    for i in range(3):
        with open(f'../data/questions/questions{i}.json', encoding='utf-8') as json_data:
            graph_json = json.load(json_data)
            all_meshes = create_node_meshes(graph_json[0]['graph']['nodes']) \
                + create_edge_meshes(graph_json[0]['graph']['links']) \
                + create_question_plane()
            graphs.append(all_meshes)

    with open('../data/relatedMeshes.json', encoding='utf-8') as json_data:
        related_meshes_matrix = json.load(json_data)
        for i in range(len(related_meshes_matrix)):
            for j in range((len(related_meshes_matrix))):
                mesh_arr = related_meshes_matrix[i][j]
                highlighted_nodes = set([m['index'] for m in mesh_arr[0]])
                task_nodes = set([node['index'] for node in mesh_arr[1]['nodes']])
                task_edges = set([f"{edge['source']['index']}-{edge['target']['index']}" for edge in mesh_arr[1]['links']])
                confusion_nodes = set([node['index'] for node in mesh_arr[2]['nodes']])
                confusion_edges = set([f"{edge['source']['index']}-{edge['target']['index']}" for edge in mesh_arr[2]['links']])
                related_meshes_matrix[i][j] = [highlighted_nodes, task_nodes, task_edges, confusion_nodes, confusion_edges]

    if save_fixations and not os.path.exists(os.path.join(dataset_dir, 'Processed')):
        os.makedirs(os.path.join(dataset_dir, 'Processed'))

    sub_dirs = [f.path for f in sorted(os.scandir(dataset_dir), key=lambda e: e.name) if f.is_dir()]
    gaze_stats_list = []
    for sub_dir in sub_dirs:
        basename = os.path.basename(sub_dir)
        if basename == 'Processed':
            continue
        participant_id = int(basename)
        if participant_id == 0: continue # 0 is discarded
        print(f'Processing participant {basename}...')
        assert os.path.exists(os.path.join(
            sub_dir, 'fixations.csv')), f"fixations.csv not found in {sub_dir}"
        assert os.path.exists(os.path.join(
            sub_dir, 'questionnaire_result.json')), f"questionnaire_result.json not found in {sub_dir}"
        assert os.path.exists(os.path.join(
            sub_dir, 'alignment.txt')), f"alignment.txt not found in {sub_dir}"
        task_ids = get_tasks(participant_id)
        df = pd.read_csv(os.path.join(sub_dir, 'fixations.csv'))
        df['start timestamp [ns]'] = np.round(df['start timestamp [ns]'] / 1e6, 3) # convert to milliseconds
        df['end timestamp [ns]'] = np.round(df['end timestamp [ns]'] / 1e6, 3)

        with open(os.path.join(sub_dir, 'alignment.txt'), encoding='utf-8') as alignment:
            timestamp_offset = np.int64(alignment.read().split('\n')[0])

        with open(os.path.join(sub_dir, 'questionnaire_result.json'), encoding='utf-8') as json_data:
            questionnaire = json.load(json_data)
        for id, task in enumerate(questionnaire['result']):
            if id < 3:  # ignore tutorials
                continue
            else:
                graph_id = math.floor(task_ids[id-3]/3)
                task_id = task_ids[id-3] % 3
                all_meshes = graphs[graph_id]
                related_meshes = related_meshes_matrix[graph_id][task_id]
                task_acc = 1 if task['answer'] == int(task['userAnswer']) else 0
                df_task = df[df['start timestamp [ns]'] +
                                timestamp_offset > np.int64(task['startTime'])]
                df_task = df_task[df_task['end timestamp [ns]'] +
                                    timestamp_offset < np.int64(task['doneTime'])]
                start_timestamp = np.int64(task['startTime'])
                complete_timestamp = np.int64(task['doneTime'])

                fix_cnts, ft_smp = process_scene(graph_id, task_id, questionnaire, df_task, timestamp_offset, plot_mesh, all_meshes, related_meshes)
                saccade_l_mean, saccade_l_std, saccade_v_mean, saccade_v_std = calc_saccade_len_vel(df_task)

                gaze_stats = {
                    'participant': participant_id,
                    'graph': graph_id,
                    'task': task_id,
                    'task_acc': task_acc,
                    'task_duration': complete_timestamp - start_timestamp,
                    'fixation_count': df_task.shape[0],
                    'saccade_len_m': saccade_l_mean,
                    'saccade_len_std': saccade_l_std,
                    'saccade_vel_m': saccade_v_mean,
                    'saccade_vel_std': saccade_v_std,
                    'fix_count_any_graph': fix_cnts[0],
                    'fix_ratio_any_graph': 0 if df_task.shape[0] == 0 else np.round(fix_cnts[0] / df_task.shape[0], 4),
                    'fix_count_question': fix_cnts[4],
                    'fix_ratio_question': 0 if fix_cnts[0] == 0 else np.round(fix_cnts[4] / fix_cnts[0], 4),
                    'fix_count_highlight': fix_cnts[1],
                    'fix_ratio_highlight': 0 if fix_cnts[0] == 0 else np.round(fix_cnts[1] / fix_cnts[0], 4),
                    'fix_count_task': fix_cnts[2],
                    'fix_ratio_task': 0 if fix_cnts[0] == 0 else np.round(fix_cnts[2] / fix_cnts[0], 4),
                    'fix_count_confusion': fix_cnts[3],
                    'fix_ratio_confusion': 0 if fix_cnts[0] == 0 else np.round(fix_cnts[3] / fix_cnts[0], 4),
                    'fix_ratio_taskall': 0 if fix_cnts[0] == 0 else np.round(fix_cnts[5]/ fix_cnts[0], 4),
                    'first_time_on_task': ft_smp["interval"]
                    # 'aoi_shift': calc_aoi_shift(df['id']),
                }
                gaze_stats_list.append(gaze_stats)

                if save_fixations:
                    if not os.path.exists(os.path.join(dataset_dir, 'Processed', f'G{str(graph_id)}', f'T{str(task_id)}')):
                        os.makedirs(os.path.join(
                            dataset_dir, 'Processed', f'G{str(graph_id)}', f'T{str(task_id)}'))
                    df_task.to_csv(os.path.join(
                        dataset_dir, 'Processed', f'G{str(graph_id)}', f'T{str(task_id)}', f'p{participant_id}.csv'), index=False)

    df_stat = pd.DataFrame.from_dict(gaze_stats_list)
    df_stat.to_csv('gaze_stats.tsv', index=False, sep='\t')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--dataset_dir", type=str,
                        default='/netpool/homes/wangyo/Dataset/NetworkGaze')
    parser.add_argument("--save_fixations", action='store_true',
                        help="Split fixations into tasks")
    parser.add_argument("--plot_mesh", action='store_true',
                        help="Split fixations into tasks")
    parser.add_argument("--analysis", action='store_true',
                        help="Assume you already have gaze_stats.tsv")
    parser.add_argument("--gaze_stats", type=str,
                        default='./gaze_stats.tsv',)
    args = vars(parser.parse_args())

    if args['analysis']:
        statistical_tests(args['gaze_stats'])
    else:
        main(args['dataset_dir'], args['save_fixations'], args['plot_mesh'])
