import os
import pandas as pd
import argparse
import matplotlib.pyplot as plt
import numpy as np
import math

def read_network_gaze_data(dataset_dir='/netpool/homes/wangyo/Dataset/NetworkGaze', G=2, T=2, p=1):
    """
    Read NetworkGaze dataset CSV file based on parameters G, T, and p.
    
    Parameters:
    -----------
    dataset_dir : str, optional
        Base directory of the dataset, default is '/netpool/homes/wangyo/Dataset/NetworkGaze'
    G : int
        Graph ID
    T : int
        Task ID
    p : int
        participant ID
    
    Returns:
    --------
    pandas.DataFrame
        DataFrame containing the data from the CSV file
    """
    file_path = os.path.join(dataset_dir, 'Processed', f'G{str(G)}', f'T{str(T)}', f'p{p}.csv')
    
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"File not found at path: {file_path}")
    
    df = pd.read_csv(file_path)
    
    print(f"Successfully loaded data with shape: {df.shape}")
    
    return df

def process_hits(hits_str):
    """Process hits string, returns list of (type, distance) tuples"""
    if pd.isna(hits_str) or hits_str == '':
        return []
    
    hits = []
    for hit in hits_str.split(';'):
        if hit and '|' in hit:  # Ensure hit is not empty and contains separator
            try:
                type_str, distance_str = hit.split('|')
                hit_type = int(type_str)
                distance = float(distance_str)
                hits.append((hit_type, distance))
            except (ValueError, IndexError):
                # Skip malformed hits
                continue

    # Sort by distance in ascending order
    hits.sort(key=lambda x: x[1], reverse=True)
    return hits

def visualize_dataframe(df, title, max_bars_per_chart=1000, bar_height=200, max_width_pixels=1000):
    """Visualize dataframe as a series of charts in a single image file with overlay line charts"""
    colors = ['lightgray', '#FF8000', '#33C060', '#50B860', '#D84545', '#DD6060', '#3075B5', '#70A5C5' ]
    labels = ['Question Panel', 'Highlighted Nodes', 'Task Nodes', 'Task Edges', 'Confusion Nodes', 'Confusion Edges', 'Other Nodes', 'Other Edges']

    # Create a copy to avoid modifying original dataframe
    df = df.copy()

    # Convert timestamps to integers
    df['start_timestamp_int'] = df['start timestamp [ns]'].astype(np.int64)
    df['end_timestamp_int'] = df['end timestamp [ns]'].astype(np.int64)

    # Process hits for all rows
    df['processed_hits'] = df['hits'].apply(process_hits)
    df = df[df['processed_hits'].apply(len) > 0]
    
    # Determine scaling factor from duration to pixels
    max_duration = df['duration [ms]'].max()
    width_scale = max_width_pixels / max_duration if max_duration > 0 else 1
    
    # Split dataframe into multiple charts
    num_rows = len(df)
    num_charts = math.ceil(num_rows / max_bars_per_chart)
    
    # Create a single large figure for all charts
    fig_height_per_chart = bar_height *1.05  # Height per chart
    total_fig_height = fig_height_per_chart * num_charts + 20  # Add extra space for legend
    fig, axs = plt.subplots(num_charts, 1, figsize=(15, total_fig_height / 100), constrained_layout=True)
    # If there's only one chart, make axs iterable
    if num_charts == 1:
        axs = [axs]
    
    # Store global maximum x value to align all charts
    global_x_max = 0

    # First pass to determine global x-axis limits
    for chart_idx in range(num_charts):
        start_idx = chart_idx * max_bars_per_chart
        end_idx = min((chart_idx + 1) * max_bars_per_chart, num_rows)
        
        chart_df = df.iloc[start_idx:end_idx].copy()
        
        if len(chart_df) == 0:
            continue

        # Determine reference timestamp for this chart
        reference_timestamp = chart_df['start_timestamp_int'].iloc[0]
        
        # Calculate position relative to reference timestamp (in milliseconds)
        chart_df['relative_start'] = chart_df['start_timestamp_int'] - reference_timestamp

        # Update global x_max if needed
        chart_x_max = chart_df['relative_start'].max() + chart_df['duration [ms]'].max() * width_scale * 1.05
        global_x_max = max(global_x_max, chart_x_max)

    # Process each chart into its own subplot
    for chart_idx in range(num_charts):
        start_idx = chart_idx * max_bars_per_chart
        end_idx = min((chart_idx + 1) * max_bars_per_chart, num_rows)

        chart_df = df.iloc[start_idx:end_idx].copy()

        if len(chart_df) == 0:
            continue

        # Get the current axis
        ax = axs[chart_idx]
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.spines['left'].set_visible(False)

        # Determine reference timestamp for this chart
        reference_timestamp = chart_df['start_timestamp_int'].iloc[0]

        # Calculate position relative to reference timestamp (in milliseconds)
        chart_df['relative_start'] = chart_df['start_timestamp_int'] - reference_timestamp

        # Set constant y-position for all bars
        y_position = bar_height
        
        # Prepare data for line charts
        x_midpoints = []
        hit_counts = []
        first_hit_ranks = []
        
        # Draw bars and collect line chart data
        for i, row in enumerate(chart_df.iterrows()):
            idx, row_data = row
            
            # Calculate bar width (pixels)
            bar_width = row_data['duration [ms]'] * width_scale
            x_position = row_data['relative_start']
            
            # Calculate midpoint for line charts
            x_midpoint = x_position + bar_width / 2
            x_midpoints.append(x_midpoint)

            # Draw main bar outline
            ax.add_patch(plt.Rectangle((x_position, y_position - bar_height/2), 
                                      bar_width, bar_height, 
                                      fill=False, edgecolor='black', linewidth=0.5))

            # Add row index text
            if i % 10 == 0 and i != 0:
                ax.text(x_position + 50, y_position + bar_height/2 + 3, 
                    f"{start_idx + i}", 
                    ha='center', va='bottom', fontsize=8)

            # Process hits for this row
            hits = row_data['processed_hits']
            
            # Count total hits
            hit_counts.append(len(hits))

            # Find rank of first hit with type 1-5
            first_hit_rank = None
            for rank, (hit_type, distance) in enumerate(hits, 1):
                if 2 <= hit_type <= 3:
                    first_hit_rank = rank
                    break
            first_hit_ranks.append(first_hit_rank)

            if hits:
                # Calculate sum of reciprocals of distances for normalization
                total_reciprocal = sum(1/dist for _, dist in hits)
                
                # Start y position from bottom of bar
                current_y = y_position + bar_height/2
                
                # Draw each hit block
                for hit_type, distance in hits:
                    # Height proportional to reciprocal of distance
                    block_height = (1/distance / total_reciprocal) * bar_height

                    # Ensure color index is valid
                    color_idx = hit_type % len(colors)

                    # Draw hit block
                    ax.add_patch(plt.Rectangle((x_position, current_y - block_height), 
                                             bar_width, block_height, 
                                             fill=True, facecolor=colors[color_idx],
                                             edgecolor='#cccccc', linewidth=0.5))

                    # Move down for next block
                    current_y -= block_height

                    
            else:
                # Fill with neutral color if no hits
                ax.add_patch(plt.Rectangle((x_position, y_position - bar_height/2), 
                                         bar_width, bar_height, 
                                         fill=True, color='#eeeeee'))

        # Add line charts if we have data
        if x_midpoints:
            # Create secondary y-axis for line charts
            ax2 = ax.twinx()
            ax2.spines['top'].set_visible(False)
            ax2.spines['bottom'].set_visible(False)
            ax2.set_yticks([])
            ax2.set_ylabel('')
            ax2.tick_params(right=False)
            ax2.spines['right'].set_visible(False)
            
            # First line chart: Number of hits
            max_hits = 1
            if hit_counts:
                max_hits = max(hit_counts) if max(hit_counts) > 0 else 1
                # Scale hit counts to fit within the bar height area
                scaled_hit_counts = [(count / max_hits) * (bar_height * 0.8) + y_position - bar_height/2 + bar_height * 0.1 
                                   for count in hit_counts]
                ax2.plot(x_midpoints, scaled_hit_counts, 'black', linewidth=1, markersize=1, 
                        alpha=0.7, label='Hit Count')

            # Second line chart: Rank of first hit
            # valid_ranks = [rank for rank in first_hit_ranks if rank is not None]
            # max_rank = 1
            # if valid_ranks:
            #     max_rank = max(valid_ranks)
            #     # Scale ranks to fit within the bar height area (inverted so rank 1 is at top)
            #     scaled_ranks = []
            #     valid_x_positions = []
            #     for i, rank in enumerate(first_hit_ranks):
            #         if rank is not None:
            #             # Invert ranking so rank 1 appears at top of scale
            #             inverted_rank = max_rank - rank + 1
            #             scaled_rank = (inverted_rank / max_rank) * (bar_height * 0.8) + y_position - bar_height/2 + bar_height * 0.1
            #             scaled_ranks.append(scaled_rank)
            #             valid_x_positions.append(x_midpoints[i])
                
            #     if scaled_ranks:
            #         ax2.plot(valid_x_positions, scaled_ranks, 'bs-', linewidth=1, markersize=1, 
            #                 alpha=0.7, label='First Hit Rank')

            # Set the same x-axis limits for the secondary axis
            if global_x_max > 60000:
                ax2.set_xlim(0, global_x_max)
            else:
                ax2.set_xlim(0, 60000)
            ax2.set_ylim(0, y_position + bar_height)

            # Add y-axis labels for line charts
            # Left y-axis (Hit Count) - position at start of x-axis
            ax.spines['left'].set_visible(True)
            ax.spines['left'].set_position(('data', 0))  # Position at x=0
            ax.yaxis.set_ticks_position('left')

            # Create hit count ticks
            hit_tick_positions = []
            hit_tick_labels = []
            for tick_value in range(0, max_hits + 1, 3): # Show up to max_hits ticks
                tick_position = (tick_value / max_hits) * (bar_height * 0.8) + y_position - bar_height/2 + bar_height * 0.1
                hit_tick_positions.append(tick_position)
                hit_tick_labels.append(str(tick_value))

            ax.set_yticks(hit_tick_positions)
            ax.set_yticklabels(hit_tick_labels)
            ax.set_ylabel('Hit Count', color='black', fontsize=10)
            ax.tick_params(axis='y', labelcolor='black', labelsize=8)

            # Right y-axis (First Hit Rank) - position at end of x-axis
            # ax2.spines['right'].set_visible(True)
            # if global_x_max > 60000:
            #     ax2.spines['right'].set_position(('data', global_x_max))  # Position at x=global_x_max
            # else:
            #     ax2.spines['right'].set_position(('data', 60000))  # Position at x=60000
            # ax2.yaxis.set_ticks_position('right')

            # Create rank ticks (inverted scale)
            # rank_tick_positions = []
            # rank_tick_labels = []
            # for rank in range(1, max_rank + 1):
            #     inverted_rank = max_rank - rank + 1
            #     tick_position = (inverted_rank / max_rank) * (bar_height * 0.8) + y_position - bar_height/2 + bar_height * 0.1
            #     rank_tick_positions.append(tick_position)
                # rank_tick_labels.append(str(rank))

            # ax2.set_yticks(rank_tick_positions)
            # ax2.set_yticklabels(rank_tick_labels)
            # ax2.set_ylabel('First Hit Rank', color='blue', fontsize=10)
            # ax2.tick_params(axis='y', labelcolor='blue', labelsize=8)
        else:
            # If no line chart data, keep original y-axis hidden
            ax.set_yticks([])

        # Set axis limits with some padding - use global x_max for consistency
        if global_x_max > 60000: # only normalize if the x-axis is extremely large
            ax.set_xlim(0, global_x_max)
        else:
            ax.set_xlim(0, 60000)
        ax.set_ylim(0, y_position + bar_height)

        # Add grid for better readability
        ax.grid(axis='x', linestyle='--', alpha=0.3)
        for ax_current in axs:
            ticks = ax_current.get_xticks()
            tick_labels = [f"{x/1000:.0f}" if x != 0 else "" for x in ticks]
            ax_current.set_xticklabels(tick_labels)

        # Set labels
        # Only add x-axis and label to bottom chart
        if chart_idx == num_charts - 1:  
            # Show the x-axis line only for the last subplot and position it at the bottom of the bars
            ax.spines['bottom'].set_visible(True)
            ax.spines['bottom'].set_position(('data', y_position - bar_height/2))
            ax.xaxis.set_ticks_position('bottom')
            ax.set_xlabel('Time (seconds)')
        else:
            ax.xaxis.set_visible(False)
        # Y-axis ticks handled in line chart section above

    # Add a single legend at the bottom of the figure (including line chart legends)
    legend_elements = [plt.Rectangle((0, 0), 1, 1, color=colors[i % len(colors)], 
                                    label=labels[i]) for i in range(8)]
    # Add line chart legend elements
    # legend_elements.extend([
    #     plt.Line2D([0], [0], color='red', marker='o', linestyle='-', 
    #                markersize=5, label='Hit Count'),
    #     plt.Line2D([0], [0], color='blue', marker='s', linestyle='-', 
    #                markersize=5, label='First Hit Rank (1-5)')
    # ])

    fig.legend(handles=legend_elements, loc='lower center', ncol=4, fontsize=9, 
               bbox_to_anchor=(0.5, -0.2), frameon=False)

    # Add overall title to the figure
    fig.suptitle(title, fontsize=12, fontweight='bold', y=0.95)

    # Save the single figure with all charts
    os.makedirs('scarf_plot', exist_ok=True)
    filename = 'scarf_plot/' + '_'.join(title.split(' ')[:6]) + '.png'
    plt.savefig(filename, dpi=300, bbox_inches='tight')
    plt.close()

    return {
        'filename': filename,
        'num_charts': num_charts,
        'total_rows': num_rows
    }

def main(args):

    data = read_network_gaze_data(
        dataset_dir=args.dataset_dir,
        G=args.G,
        T=args.T,
        p=args.p
    )

    visualize_dataframe(data, f'Participant {args.p} Graph {args.G} Task {args.T}')


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Read NetworkGaze dataset CSV files')
    
    parser.add_argument('--dataset_dir', type=str, 
                        default='/netpool/homes/wangyo/Dataset/NetworkGaze', # "D:\\webxr\\study"
                        help='Base directory of the dataset')
    parser.add_argument('--G', type=str, default=2,
                        help='Graph ID')
    parser.add_argument('--T', type=str, default=2,
                        help='Task ID')
    parser.add_argument('--p', type=str, default=1,
                        help='participant ID')

    args = parser.parse_args()
    main(args)
    # for g in range(3):
        # for t in range(3):
        #     for p in range(1,19):
        #         data = read_network_gaze_data(args.dataset_dir, g, t, p)
        #         visualize_dataframe(data, f'Participant {p} Graph {g} Task {t} Timeline Visualization')
