import numpy as np
import pandas as pd
import statsmodels.api as sm
import statsmodels.formula.api as smf
from scipy.stats import spearmanr
from statsmodels.stats.multicomp import pairwise_tukeyhsd

def get_tasks(batch_id: int):
    batch_id = int(batch_id % 9)-1
    TASKS = [8, 1, 5, 6, 2, 3, 7, 0, 4]
    # print(TASKS[batch_id:] + TASKS[:batch_id])
    return TASKS[batch_id:] + TASKS[:batch_id]

def update_fix_cnt(ft_smp, fix_flag, camera_loc, hit_id, related_meshes, task_id, hits_stat, distance):
    """ update the fixation count and first time on task from hited meshes
    """
    fix_flag[0] = True
    
    if (task_id == 0 or task_id == 2) and hit_id in related_meshes[0] :
        if not ft_smp["flag"]:
            ft_smp["flag"] = True
            ft_smp["interval"] = np.int64(camera_loc['timeStamp']) - ft_smp['start']
    if task_id == 1 and (hit_id in related_meshes[1]):
        if not ft_smp["flag"]:
            ft_smp["flag"] = True
            ft_smp["interval"] = np.int64(camera_loc['timeStamp']) - ft_smp['start']
    
    if hit_id in related_meshes[0]:
        fix_flag[1] = True
        hits_stat.append({"type": 1, "distance": distance})
    # task nodes
    elif hit_id in related_meshes[1]: 
        fix_flag[2] = True
        hits_stat.append({"type": 2, "distance": distance})
    # task edges
    elif hit_id in related_meshes[2]:
        fix_flag[2] = True
        hits_stat.append({"type": 3, "distance": distance})
    # confusion nodes
    elif hit_id in related_meshes[3]: 
        fix_flag[3] = True
        hits_stat.append({"type": 4, "distance": distance})
    # confusion edges
    elif hit_id in related_meshes[4]:
        fix_flag[3] = True
        hits_stat.append({"type": 5, "distance": distance})
    # non task nodes
    elif isinstance(hit_id, int):
        hits_stat.append({"type": 6, "distance": distance})
    # non task edges
    elif isinstance(hit_id, str) :
        hits_stat.append({"type": 7, "distance": distance})


def find_camera_pose(questionnaire, start_time, end_time):
    """ find the camera pose in the questionnaire
    Args:
        questionnaire (json): questionnaire and camera poses
        start_time (int): start time of the task
        end_time (int): end time of the task
    Return:
        camera pose (json)
    """
    for _, cameraLoc in enumerate(questionnaire['cameralLocations']):
        if np.int64(cameraLoc['timeStamp']) >= start_time and np.int64(cameraLoc['timeStamp']) <= end_time:
            return cameraLoc
    return None


def two_way_anova_with_tukey(df_stat: pd.DataFrame, dependent_var: str, group1: str, group2: str):
    """
    Perform two-way ANOVA with Tukey's HSD post-hoc test.

    Args:
        df_stat (pd.DataFrame): DataFrame containing the data.
        dependent_var (str): The dependent variable for the ANOVA.
        group1 (str): The first grouping variable.
        group2 (str): The second grouping variable.

    Returns:
        None
    """
    model = smf.ols(f'{dependent_var} ~ C({group1}) * C({group2})', data=df_stat).fit()
    anova_table = sm.stats.anova_lm(model, typ=2)
    print(f'{dependent_var} ~ C({group1}) * C({group2})\n', anova_table)

    # Run Tukey HSD, pairwise post-hoc test
    df_stat['interaction'] = df_stat[group1].astype(str) + "_" + df_stat[group2].astype(str)
    tukey = pairwise_tukeyhsd(endog=df_stat[dependent_var],
                            groups=df_stat['interaction'],
                            alpha=0.05)
    print(f'Tukey {dependent_var} ~ C({group1}) * C({group2})\n', tukey)
    # Optional: plot the confidence intervals
    # tukey.plot_simultaneous()

def basic_stat(df_stat: pd.DataFrame, var: str, factor: float = 1, precision = 2) -> None:
    """
    Perform basic statistical analysis on the given DataFrame.

    Args:
        df_stat (pd.DataFrame): DataFrame containing the data.
    """
    means_all = df_stat[var].mean()
    stds_all = df_stat[var].std()
    print(f"Means of {var}:\n", np.round(means_all/factor, precision))
    print("Standard Deviations:\n", np.round(stds_all/factor, precision))
    means_task = df_stat.groupby(['task'])[var].mean()
    # stds_task = df_stat.groupby(['task'])[var].std()
    print(f"Means of {var} by task:\n", np.round(means_task/factor, precision))
    # print("Standard Deviations:\n", np.round(stds_task/factor, precision))
    means_graph = df_stat.groupby(['graph'])[var].mean()
    # stds_graph = df_stat.groupby(['graph'])[var].std()
    print(f"Means of {var} by graph:\n", np.round(means_graph/factor, precision))
    # print("Standard Deviations:\n", np.round(stds_graph/factor, precision))
    means = df_stat.groupby(['graph', 'task'])[var].mean()
    # stds = df_stat.groupby(['graph', 'task'])[var].std()
    print(f"Means of {var}:\n", np.round(means/factor, precision))
    # print("Standard Deviations:\n", np.round(stds/factor, precision))

def spearman_corr(df_stat: pd.DataFrame, var1: str, var2: str) -> None:
    for task_id in range(3):
        corr, p_value = spearmanr(df_stat[df_stat['task']==task_id][var1], df_stat[df_stat['task']==task_id][var2])
        print(f"Spearman correlation between {var1} and {var2}, task {task_id}: {corr:.4f}, P-value: {p_value:.4g}")

def do_tests(df_stat: pd.DataFrame) -> None:
    # some mean and stds
    basic_stat(df_stat, 'task_duration', factor=1000, precision=2)
    basic_stat(df_stat, 'task_acc', factor=0.01, precision=2) # percentage
    basic_stat(df_stat, 'saccade_vel_m', precision=4) # percentage
    basic_stat(df_stat, 'saccade_len_m', precision=2)
    # basic_stat(df_stat, 'fixation_count', precision=2)
    # basic_stat(df_stat, 'fix_count_task', precision=2)
    basic_stat(df_stat, 'fix_ratio_task', precision=4)

    # Spearman correlation coefficients
    spearman_corr(df_stat, 'fix_ratio_task', 'task_acc')
    spearman_corr(df_stat, 'saccade_len_m', 'task_acc')
    spearman_corr(df_stat, 'saccade_vel_m', 'task_acc')
    spearman_corr(df_stat, 'task_duration', 'task_acc')
    # spearman_corr(df_stat, 'fix_ratio_taskall', 'task_acc')

def statistical_tests(stat_path: str) -> None:
    """
    Perform statistical tests on the given DataFrame.

    Args:
        gaze_path: path of the DataFrame
    """
    df_stat = pd.read_csv(stat_path, sep='\t')
    df_stat['graph'] = df_stat['graph'].astype('category')
    df_stat['task'] = df_stat['task'].astype('category')

    do_tests(df_stat)

    # test interaction effect of tasks and graphs
    # Ensure categorical variables are treated as categorical

    # Three-way ANOVA with interaction, found that task_acc is not a significant factor for fix_ratio_task, and fix_ratio_taskall
    # df_stat['task_acc'] = df_stat['task_acc'].astype('category')
    # model = smf.ols('fix_ratio_task ~ C(graph) * C(task) * C(task_acc)', data=df_stat).fit()
    # model = smf.ols('task_acc ~ C(graph) * C(task) + fix_ratio_taskall', data=df_stat).fit()
    # anova_table = sm.stats.anova_lm(model, typ=2)
    # print(anova_table)

    # two_way_anova_with_tukey(df_stat, 'task_duration', 'graph', 'task')
    # two_way_anova_with_tukey(df_stat, 'fixation_count', 'graph', 'task')
    two_way_anova_with_tukey(df_stat, 'saccade_vel_m', 'graph', 'task')
    two_way_anova_with_tukey(df_stat, 'saccade_len_m', 'graph', 'task')
    two_way_anova_with_tukey(df_stat, 'fix_ratio_task', 'graph', 'task')
    # two_way_anova_with_tukey(df_stat, 'fix_count_task', 'graph', 'task')
    # two_way_anova_with_tukey(df_stat, 'fix_ratio_highlight', 'graph', 'task')
    # two_way_anova_with_tukey(df_stat, 'fix_ratio_taskall', 'graph', 'task')
