// Import Babylon.js and GUI modules
import * as BABYLON from '@babylonjs/core'
import * as gui from '@babylonjs/gui'
import draw from './draw'
import replayDraw from './replay/draw'
import questions0 from './data/questions/questions0.json'
import questions1 from './data/questions/questions1.json'
import questions2 from './data/questions/questions2.json'
import questionsTT0 from './data/questions/questionsTutorial0.json'
import questionsTT1 from './data/questions/questionsTutorial1.json'
import questionsTT2 from './data/questions/questionsTutorial2.json'

export default function createQuestionnaire (
  scene,
  buttonClickingHistory,
  cameralLocations,
  calibrationTimeStamps,
  task
) {
  const questionSet = [...questions0, ...questions1, ...questions2]

  const path = window.location.pathname
  let selectedQuestionIndexes

  switch (task || path.at(-2)) {
    case '2':
      selectedQuestionIndexes = [1, 5, 6, 2, 3, 7, 0, 4, 8]
      break
    case '3':
      selectedQuestionIndexes = [5, 6, 2, 3, 7, 0, 4, 8, 1]
      break
    case '4':
      selectedQuestionIndexes = [6, 2, 3, 7, 0, 4, 8, 1, 5]
      break
    case '5':
      selectedQuestionIndexes = [2, 3, 7, 0, 4, 8, 1, 5, 6]
      break
    case '6':
      selectedQuestionIndexes = [3, 7, 0, 4, 8, 1, 5, 6, 2]
      break
    case '7':
      selectedQuestionIndexes = [7, 0, 4, 8, 1, 5, 6, 2, 3]
      break
    case '8':
      selectedQuestionIndexes = [0, 4, 8, 1, 5, 6, 2, 3, 7]
      break
    case '9':
      selectedQuestionIndexes = [4, 8, 1, 5, 6, 2, 3, 7, 0]
      break
    default:
      selectedQuestionIndexes = [8, 1, 5, 6, 2, 3, 7, 0, 4]
      break
  }

  const questions = [
    questionsTT0[0],
    questionsTT2[1],
    questionsTT1[2],
    questionSet[selectedQuestionIndexes[0]],
    questionSet[selectedQuestionIndexes[1]],
    questionSet[selectedQuestionIndexes[2]],
    questionSet[selectedQuestionIndexes[3]],
    questionSet[selectedQuestionIndexes[4]],
    questionSet[selectedQuestionIndexes[5]],
    questionSet[selectedQuestionIndexes[6]],
    questionSet[selectedQuestionIndexes[7]],
    questionSet[selectedQuestionIndexes[8]]
  ]

  // Create a plane for the questionnaire
  const plane = BABYLON.MeshBuilder.CreatePlane(
    'questionnairePlane',
    { width: 0.5, height: 1 },
    scene
  )
  plane.position = new BABYLON.Vector3(0, 0.15, -1.5)
  plane.rotation.y = Math.PI / -1

  // Create AdvancedDynamicTexture for the GUI
  const advancedTexture = gui.AdvancedDynamicTexture.CreateForMesh(
    plane,
    1024,
    2048
  )
  advancedTexture.background = '#86a6bf'

  let CoT,
    buttonHandlers = {}

  let currentQuestionIndex = 0

  // Function to update the question and progress
  const updateQuestion = () => {
    const question = questions[currentQuestionIndex]
    CoT && CoT.dispose()

    // Clear previous GUI elements
    const children = [...advancedTexture.getChildren()] // Copy the array to avoid modification during iteration
    children.forEach(child => {
      advancedTexture.removeControl(child) // Detach the control
      child.dispose() // Dispose of the control to free resources
    })

    // Display progress
    const progressRect = new gui.Rectangle()
    progressRect.width = '800px'
    progressRect.height = '200px'
    progressRect.thickness = 0 // No border
    progressRect.background = 'white' // White background
    progressRect.horizontalAlignment = gui.Control.HORIZONTAL_ALIGNMENT_CENTER
    progressRect.verticalAlignment = gui.Control.VERTICAL_ALIGNMENT_BOTTOM
    progressRect.paddingBottom = '60px'

    const progressText = new gui.TextBlock()
    progressText.text = `${currentQuestionIndex + 1} / ${questions.length}`
    progressText.fontSize = '80px' // Adjust font size for better fit
    progressText.color = 'black' // Black text
    progressText.textHorizontalAlignment =
      gui.Control.HORIZONTAL_ALIGNMENT_CENTER
    progressText.textVerticalAlignment = gui.Control.VERTICAL_ALIGNMENT_CENTER

    progressRect.addControl(progressText) // Add the text to the rectangle
    advancedTexture.addControl(progressRect) // Add the rectangle to the texture

    const questionText = showQuestionText(question)
    showStartInfo(question, questionText)
  }

  const showQuestionText = question => {
    // Display question text
    const questionText = new gui.TextBlock()
    questionText.text = question.text
    questionText.fontSize = '80px'
    questionText.color = 'white'
    questionText.textWrapping = true
    questionText.textHorizontalAlignment =
      gui.Control.HORIZONTAL_ALIGNMENT_CENTER
    questionText.textVerticalAlignment = gui.Control.VERTICAL_ALIGNMENT_TOP
    questionText.paddingTop = '60px'
    advancedTexture.addControl(questionText)

    return questionText
  }

  const showStartInfo = (question, questionText) => {
    const startPanel = new gui.Grid()
    startPanel.height = '80%'
    startPanel.horizontalAlignment = gui.Control.HORIZONTAL_ALIGNMENT_CENTER
    startPanel.verticalAlignment = gui.Control.VERTICAL_ALIGNMENT_TOP
    startPanel.paddingTop = '400px'
    advancedTexture.addControl(startPanel)
    startPanel.addColumnDefinition(1)
    startPanel.addRowDefinition(1)
    const startButton = gui.Button.CreateSimpleButton('start', 'start task')
    startButton.fontSize = '80px'
    startButton.width = '400px'
    startButton.height = '200px'
    startButton.color = 'black'
    startButton.background = '#edb50a'

    const handleStart = () => {
      question.startTime = Date.now()

      showDoneInfo(question, questionText)
      CoT = task ? replayDraw(question, scene) : draw(question, scene)

      buttonClickingHistory.key = 'start'
      buttonClickingHistory.index++

      advancedTexture.removeControl(startPanel)
      startPanel.dispose()
    }
    startButton.onPointerClickObservable.add(handleStart)

    buttonHandlers.start = handleStart
    startPanel.addControl(startButton, 0, 0)
  }

  const showDoneInfo = (question, questionText) => {
    const donePanel = new gui.Grid()
    donePanel.height = '80%'
    donePanel.horizontalAlignment = gui.Control.HORIZONTAL_ALIGNMENT_CENTER
    donePanel.verticalAlignment = gui.Control.VERTICAL_ALIGNMENT_TOP
    donePanel.paddingTop = '400px'
    advancedTexture.addControl(donePanel)

    donePanel.addColumnDefinition(1)
    donePanel.addRowDefinition(2)

    const hintText = new gui.TextBlock()
    hintText.text = 'The graph is displayed behind you.'
    hintText.fontSize = '60px'
    hintText.color = 'white'
    hintText.textWrapping = true
    hintText.textHorizontalAlignment = gui.Control.HORIZONTAL_ALIGNMENT_CENTER
    hintText.textVerticalAlignment = gui.Control.VERTICAL_ALIGNMENT_TOP
    hintText.paddingTop = '200px'
    donePanel.addControl(hintText, 0, 0)

    const doneButton = gui.Button.CreateSimpleButton('done', 'done')
    doneButton.fontSize = '80px'
    doneButton.width = '400px'
    doneButton.height = '200px'
    doneButton.color = 'black'
    doneButton.background = '#edb50a'

    const handleDone = () => {
      question.doneTime = Date.now()

      showQuestion(question, questionText)

      buttonClickingHistory.key = 'done'
      buttonClickingHistory.index++

      advancedTexture.removeControl(donePanel)
      donePanel.dispose()
    }
    doneButton.onPointerClickObservable.add(handleDone)

    buttonHandlers.done = handleDone
    donePanel.addControl(doneButton, 1, 0)
  }

  const showQuestion = (question, questionText) => {
    // Add answer buttons or input
    const buttonPanel = new gui.Grid()
    buttonPanel.height = '80%'
    buttonPanel.horizontalAlignment = gui.Control.HORIZONTAL_ALIGNMENT_CENTER
    buttonPanel.verticalAlignment = gui.Control.VERTICAL_ALIGNMENT_TOP
    buttonPanel.paddingTop = '400px'
    advancedTexture.addControl(buttonPanel)

    if (
      question.answers.length === 1 &&
      question.answers[0] === 'Enter node label'
    ) {
      buttonPanel.addColumnDefinition(0.3)
      buttonPanel.addColumnDefinition(0.3)
      buttonPanel.addColumnDefinition(0.3)
      buttonPanel.addRowDefinition(1)
      buttonPanel.addRowDefinition(1)
      buttonPanel.addRowDefinition(1)
      buttonPanel.addRowDefinition(1)
      buttonPanel.addRowDefinition(1)

      // Create the input field
      const inputField = new gui.InputText()
      inputField.fontSize = '80px'
      inputField.width = '300px'
      inputField.height = '200px'
      inputField.color = 'black'
      inputField.background = 'white'
      inputField.text = ''
      inputField.isReadOnly = true // Only virtual input allowed
      buttonPanel.addControl(inputField, 0, 0) // Add to the first row, first column

      // Create the delete button
      const deleteButton = gui.Button.CreateSimpleButton('Delete', '⌫')
      deleteButton.fontSize = '80px'
      deleteButton.width = '300px'
      deleteButton.height = '200px'
      deleteButton.color = 'black'
      deleteButton.background = '#edb50a'

      const handleDelete = () => {
        inputField.text = inputField.text.slice(0, -1) // Remove the last character

        buttonClickingHistory.key = 'delete'
        buttonClickingHistory.index++
      }
      deleteButton.onPointerClickObservable.add(handleDelete)
      buttonHandlers.delete = handleDelete
      buttonPanel.addControl(deleteButton, 0, 2) // Add to the first row, third column

      // Numeric keys (1-9)
      const numericKeys = [
        ['1', '2', '3'],
        ['4', '5', '6'],
        ['7', '8', '9']
      ]
      numericKeys.forEach((row, rowIndex) => {
        row.forEach((key, colIndex) => {
          const keyButton = gui.Button.CreateSimpleButton(`key_${key}`, key)
          keyButton.fontSize = '80px'
          keyButton.width = '300px'
          keyButton.height = '200px'
          keyButton.color = 'black'
          keyButton.background = '#edb50a'
          const handleKey = () => {
            inputField.text += key // Append the key to the input field

            buttonClickingHistory.key = key
            buttonClickingHistory.index++
          }
          keyButton.onPointerClickObservable.add(handleKey)

          buttonHandlers[key] = handleKey
          buttonPanel.addControl(keyButton, rowIndex + 1, colIndex) // Add to respective row and column
        })
      })

      // Add the "0" button
      const zeroButton = gui.Button.CreateSimpleButton('key_0', '0')
      zeroButton.fontSize = '80px'
      zeroButton.width = '300px'
      zeroButton.height = '200px'
      zeroButton.color = 'black'
      zeroButton.background = '#edb50a'

      const hadleZero = () => {
        inputField.text += '0' // Append "0" to the input field

        buttonClickingHistory.key = '0'
        buttonClickingHistory.index++
      }
      zeroButton.onPointerClickObservable.add(hadleZero)

      buttonHandlers['0'] = hadleZero
      buttonPanel.addControl(zeroButton, 4, 1) // Add to fifth row, 2nd column

      // Add the submit button
      const submitButton = gui.Button.CreateSimpleButton('Submit', 'OK')
      submitButton.fontSize = '80px'
      submitButton.width = '300px'
      submitButton.height = '200px'
      submitButton.color = 'black'
      submitButton.background = '#edb50a'

      const handleSubmit = () => {
        if (inputField.text.trim() !== '') {
          showAnswer(question, questionText, inputField.text.trim())

          buttonClickingHistory.key = 'submit'
          buttonClickingHistory.index++

          advancedTexture.removeControl(buttonPanel)
          buttonPanel.dispose()
        }
      }
      submitButton.onPointerClickObservable.add(handleSubmit)

      buttonHandlers.submit = handleSubmit
      buttonPanel.addControl(submitButton, 4, 2) // Add to fifth row, third column
    } else {
      buttonPanel.addColumnDefinition(0.5)
      buttonPanel.addColumnDefinition(0.5)
      buttonPanel.addRowDefinition(0.5)
      buttonPanel.addRowDefinition(0.5)
      buttonPanel.addRowDefinition(0.5)
      buttonPanel.addRowDefinition(0.5)
      buttonPanel.addRowDefinition(0.5)

      question.answers.forEach((answer, index) => {
        const button = gui.Button.CreateSimpleButton('answer', answer)
        button.fontSize = '60px'
        button.width = '300px'
        button.height = '150px'
        button.color = 'black'
        button.background = '#edb50a'

        const handleKey = () => {
          showAnswer(question, questionText, answer)

          buttonClickingHistory.key = answer
          buttonClickingHistory.index++

          advancedTexture.removeControl(buttonPanel)
          buttonPanel.dispose()
        }
        button.onPointerClickObservable.add(handleKey)
        buttonHandlers[answer] = handleKey
        buttonPanel.addControl(button, Math.floor(index / 2), index % 2)
      })
    }
  }

  const showAnswer = (question, questionText, answer) => {
    advancedTexture.removeControl(questionText)
    questionText.dispose()

    const answerText = new gui.TextBlock()
    answerText.text = 'You choose answer:\n' + answer + '\nIs that correct?'
    answerText.fontSize = '80px'
    answerText.color = 'white'
    answerText.textWrapping = true
    answerText.textHorizontalAlignment = gui.Control.HORIZONTAL_ALIGNMENT_CENTER
    answerText.textVerticalAlignment = gui.Control.VERTICAL_ALIGNMENT_TOP
    answerText.paddingTop = '60px'
    advancedTexture.addControl(answerText)

    const answerPanel = new gui.Grid()
    answerPanel.height = '80%'
    answerPanel.horizontalAlignment = gui.Control.HORIZONTAL_ALIGNMENT_CENTER
    answerPanel.verticalAlignment = gui.Control.VERTICAL_ALIGNMENT_TOP
    answerPanel.paddingTop = '400px'
    advancedTexture.addControl(answerPanel)
    answerPanel.addColumnDefinition(0.5)
    answerPanel.addColumnDefinition(0.5)
    answerPanel.addRowDefinition(1)

    const nextButton = gui.Button.CreateSimpleButton('next', 'yes,\nnext task')
    nextButton.fontSize = '80px'
    nextButton.width = '400px'
    nextButton.height = '200px'
    nextButton.color = 'black'
    nextButton.background = '#edb50a'

    const handleNext = download => {
      question.userAnswer = answer

      buttonClickingHistory.key = 'next'
      buttonClickingHistory.index++

      advancedTexture.removeControl(answerText)
      answerText.dispose()
      advancedTexture.removeControl(answerPanel)
      answerPanel.dispose()
      goToNextQuestion(download)
    }
    nextButton.onPointerClickObservable.add(handleNext)
    buttonHandlers.next = handleNext
    answerPanel.addControl(nextButton, 0, 1)

    const backButton = gui.Button.CreateSimpleButton('back', 'no,\ngo back')
    backButton.fontSize = '80px'
    backButton.width = '400px'
    backButton.height = '200px'
    backButton.color = 'black'
    backButton.background = '#edb50a'

    const handleBack = () => {
      const questionText = showQuestionText(question)
      showQuestion(question, questionText)

      buttonClickingHistory.key = 'back'
      buttonClickingHistory.index++

      advancedTexture.removeControl(answerText)
      answerText.dispose()
      advancedTexture.removeControl(answerPanel)
      answerPanel.dispose()
    }
    backButton.onPointerClickObservable.add(handleBack)
    buttonHandlers.back = handleBack
    answerPanel.addControl(backButton, 0, 0)
  }

  // Function to go to the next question
  const goToNextQuestion = download => {
    currentQuestionIndex++
    if (currentQuestionIndex < questions.length) {
      updateQuestion()
    } else {
      showCompletionMessage(download)
    }
  }

  // Function to show the completion message
  const showCompletionMessage = (download = true) => {
    // Clear previous GUI elements
    const children = [...advancedTexture.getChildren()] // Copy the array to avoid modification during iteration
    children.forEach(child => {
      advancedTexture.removeControl(child) // Detach the control
      child.dispose() // Dispose of the control to free resources
    })

    const completionText = new gui.TextBlock()
    completionText.text = 'Thank you for completing the questionnaire!'
    completionText.textWrapping = true
    completionText.fontSize = '100px'
    completionText.color = 'white'
    completionText.textHorizontalAlignment =
      gui.Control.HORIZONTAL_ALIGNMENT_CENTER
    completionText.textVerticalAlignment = gui.Control.VERTICAL_ALIGNMENT_CENTER
    advancedTexture.addControl(completionText)

    if (!download) {
      return
    }

    const result = questions.map(q => {
      return {
        question: q.question,
        layers: q.layers,
        layout: q.layout,
        answer: q.answer,
        userAnswer: q.userAnswer,
        startTime: q.startTime,
        doneTime: q.doneTime
      }
    })

    const resultJson = JSON.stringify({
      task: path.at(-2),
      calibrationTimeStamps,
      result,
      cameralLocations
    })

    const blob = new Blob([resultJson], { type: 'application/json' })
    const url = URL.createObjectURL(blob)

    const a = document.createElement('a')
    a.href = url
    a.download = 'questionnaire_result'
    document.body.appendChild(a)
    a.click()

    document.body.removeChild(a)
    URL.revokeObjectURL(url)
  }

  // Initialize the first question
  updateQuestion()

  return buttonHandlers
}
