import * as anu from '@jpmorganchase/anu'
import * as BABYLON from '@babylonjs/core'

export default function draw (question, scene) {
  const data = question.graph
  console.log(data)

  const highlightedNodes = new Set(question.highlightedNodes)

  //We use Mesh instancing here for better performance, first we create a Mesh that serves as the root Node
  let rootSphere = anu.create('sphere', 'node', { diameter: 0.01 })
  rootSphere.isVisible = false
  rootSphere.registerInstancedBuffer('color', 4) //We need an InstancedBuffer to set the color of instances

  //Create a Center of Transform TransformNode that serves the parent node for all our meshes that make up our chart
  let CoT = anu.create('cot', 'cot')

  rootSphere.parent = CoT

  //Select our CoT so that we have it as a Selection object
  let network = anu.selectName('cot', scene)

  //Create instanced sphere meshes from our rootSphere as children of our CoT for each row of our data and set their initial visual encodings using method chaining
  let nodes = network
    .bindInstance(rootSphere, data.nodes)
    .position(d => new BABYLON.Vector3(d.x, d.y, d.z))
    .setInstancedBuffer('color', d =>
      highlightedNodes.has(d.index)
        ? new BABYLON.Color4(1, 0, 0, 1)
        : new BABYLON.Color4(1, 1, 0, 1)
    )
    .id(d => d.index)

  //Create a helper function that will return us an array of arrays where each sub-array is the start and end Vector3 of each link
  function dataToLinks (v) {
    let start = new BABYLON.Vector3(v.source.x, v.source.y, v.source.z)
    let end = new BABYLON.Vector3(v.target.x, v.target.y, v.target.z)

    return [start, end]
  }

  //Create our lineSystem mesh using our data and helper function from above
  let links = network
    .bind(
      'lines',
      { points: d => dataToLinks(d), updatable: true },
      data.links
    )
    .name(d => `Link ${d.source.index}-${d.target.index}`)
    .prop('color', new BABYLON.Color4(1, 1, 1, 1))

  let labels = createLabelsWithSPS(scene, data.nodes, highlightedNodes)
  labels.forEach(l => (l.parent = CoT))

  return CoT
}

function createLabelsWithSPS (scene, nodes, highlightedNodes) {
  const labelSizePx = 128
  const fontSizePx = 40
  const font = `bold ${fontSizePx}px monospace`
  const labelSize = 0.05

  const maxTextureSizePx = 1280
  const maxLabelsPerDimension = Math.floor(maxTextureSizePx / labelSizePx)
  const maxLabelsPerTexture = Math.pow(maxLabelsPerDimension, 2)

  const materials = []
  let textures = []
  const labelPlanes = []

  const numTextures = Math.ceil(nodes.length / maxLabelsPerTexture)

  // create textures and materials
  for (let i = 0; i < numTextures; i++) {
    const texture = new BABYLON.DynamicTexture(
      'label-texture' + i,
      { width: maxTextureSizePx, height: maxTextureSizePx },
      scene,
      false
    )

    textures.push(texture)

    const material = new BABYLON.StandardMaterial('label-material' + i, scene)
    material.diffuseTexture = texture
    material.diffuseTexture.hasAlpha = true
    material.specularColor = new BABYLON.Color3(0, 0, 0)
    material.emissiveColor = new BABYLON.Color3(1, 1, 1)
    material.backFaceCulling = false

    materials.push(material)
  }

  // crete label planes
  for (let i = 0; i < nodes.length; i++) {
    const node = nodes[i]

    // which texture to use for nodes
    const textureToUse = Math.floor(i / maxLabelsPerTexture)
    const indexInTexture = i % maxLabelsPerTexture

    // draw text on texture
    const labelPos = drawLabelText(
      textures[textureToUse],
      node.index.toString(),
      indexInTexture,
      maxLabelsPerDimension,
      labelSizePx,
      fontSizePx,
      font,
      highlightedNodes
    )

    const labelPlane = BABYLON.MeshBuilder.CreatePlane(
      'label_' + node.index,
      { width: labelSize, height: labelSize, updatable: true },
      scene
    )

    labelPlane.position.x = node.x - 0.02
    labelPlane.position.y = node.y + 0.01
    labelPlane.position.z = node.z
    // labelPlane.billboardMode = 7

    // set UV coordinates
    labelPlane.material = materials[textureToUse]

    // get and update UV coordinates
    const uvs = labelPlane.getVerticesData(BABYLON.VertexBuffer.UVKind)

    const uMin = labelPos.left / maxTextureSizePx
    const uMax = labelPos.right / maxTextureSizePx
    const vMax = 1 - labelPos.top / maxTextureSizePx
    const vMin = 1 - labelPos.bottom / maxTextureSizePx
  
    // UV plane coordinates
    uvs[0] = uMin // bottom left
    uvs[1] = vMin // bottom left
    uvs[2] = uMax
    uvs[3] = vMin
    uvs[4] = uMax // top right
    uvs[5] = vMax // top right
    uvs[6] = uMin
    uvs[7] = vMax

    // mesh UV
    labelPlane.updateVerticesData(BABYLON.VertexBuffer.UVKind, uvs)

    labelPlanes.push(labelPlane)
  }

  // 更新所有纹理
  for (let i = 0; i < textures.length; i++) {
    textures[i].update()
  }

  return labelPlanes
}

// drawLabelText
function drawLabelText (
  texture,
  label,
  index,
  maxLabelsPerDimension,
  labelSizePx,
  fontSizePx,
  font,
  highlightedNodes
) {
  // text row and column in texture
  const row = Math.floor(index / maxLabelsPerDimension)
  const col = index % maxLabelsPerDimension

  // text coordinates in texture
  const left = labelSizePx * col
  const right = left + labelSizePx
  const top = labelSizePx * row
  const bottom = top + labelSizePx

  const textureContext = texture.getContext()
  // transparent background
  textureContext.clearRect(left, top, labelSizePx, labelSizePx)

  texture.drawText(
    label,
    left + labelSizePx / 2, // X center of text
    top + labelSizePx / 2, // Y center
    font,
    highlightedNodes.has(+label) ? 'red' : 'yellow',
    'transparent',
    true, // horizontal center
    true // vertical center
  )
  return { left, right, top, bottom }
}
