import '@babylonjs/core/Debug/debugLayer'
import '@babylonjs/inspector'
import { Engine } from '@babylonjs/core'
import Papa from 'papaparse'
import { createScene } from './renderScene.js'

//Grab DOM element where we will attach our canvas. #app is the id assigned to an empty <div> in our index.html
const app = document.querySelector('#app')

//Create a canvas element and append it to #app div
const form = document.createElement('form')
form.id = 'uploadForm'

// Create file input
const logInput = document.createElement('input')
logInput.type = 'file'
logInput.id = 'LogInput'

const fixationsInput = document.createElement('input')
fixationsInput.type = 'file'
fixationsInput.id = 'fixations'

const timesetOffsetLabel = document.createElement('label')
timesetOffsetLabel.textContent = 'Timeset Offset: '
const timesetOffsetInput = document.createElement('input')
timesetOffsetInput.type = 'number'
timesetOffsetInput.id = 'timesetOffset'
timesetOffsetInput.value = '0'
timesetOffsetInput.style.marginRight = '15px'

const leftOffsetLabel = document.createElement('label')
leftOffsetLabel.textContent = 'Left Offset: '
const leftOffsetInput = document.createElement('input')
leftOffsetInput.type = 'number'
leftOffsetInput.id = 'leftOffset'
leftOffsetInput.value = '770'
leftOffsetInput.style.marginRight = '15px'

const topOffsetLabel = document.createElement('label')
topOffsetLabel.textContent = 'Top Offset: '
const topOffsetInput = document.createElement('input')
topOffsetInput.type = 'number'
topOffsetInput.id = 'topOffset'
topOffsetInput.value = '610'

const replayButton = document.createElement('button')
replayButton.type = 'button'
replayButton.textContent = 'Replay'
replayButton.addEventListener('click', replay)

form.appendChild(logInput)
form.appendChild(fixationsInput)
form.appendChild(timesetOffsetLabel)
form.appendChild(timesetOffsetInput)
form.appendChild(leftOffsetLabel)
form.appendChild(leftOffsetInput)
form.appendChild(topOffsetLabel)
form.appendChild(topOffsetInput)
form.appendChild(replayButton)

app.appendChild(form)

//initialize babylon engine, passing in our target canvas element, and create a new scene
const canvas = document.createElement('canvas')
app.appendChild(canvas)
const babylonEngine = new Engine(canvas, true)

//create a scene object using our engine
let scene

//Listen for window size changes and resize the scene accordingly
window.addEventListener('resize', function () {
  babylonEngine.resize()
})

// hide/show the Inspector
window.addEventListener('keydown', ev => {
  // Shift+Ctrl+Alt+I
  if (ev.shiftKey && ev.ctrlKey && ev.altKey && ev.keyCode === 73) {
    if (scene.debugLayer.isVisible()) {
      scene.debugLayer.hide()
    } else {
      scene.debugLayer.show()
    }
  }
})

function replay () {
  const logFile = logInput.files[0]
  const fixationsFile = fixationsInput.files[0]

  // 获取三个输入框的值
  const timesetOffset =
    parseFloat(document.getElementById('timesetOffset').value) || 0
  const leftOffset =
    parseFloat(document.getElementById('leftOffset').value) || 770
  const topOffset =
    parseFloat(document.getElementById('topOffset').value) || 610

  // 创建两个 Promise 分别处理两个文件的读取
  const readLogFile = new Promise((resolve, reject) => {
    const logReader = new FileReader()
    logReader.onload = function (event) {
      try {
        const log = JSON.parse(event.target.result)
        resolve(log)
      } catch (error) {
        reject(error)
      }
    }
    logReader.onerror = reject
    logReader.readAsText(logFile)
  })

  const readFixationsFile = new Promise((resolve, reject) => {
    const fixationsReader = new FileReader()
    fixationsReader.onload = function (event) {
      try {
        // 将三个偏移值传递给 readCSVFromString 函数
        const fixations = readCSVFromString(
          event.target.result,
          timesetOffset,
          leftOffset,
          topOffset
        )
        resolve(fixations)
      } catch (error) {
        reject(error)
      }
    }
    fixationsReader.onerror = reject
    fixationsReader.readAsText(fixationsFile)
  })

  // 等待两个文件都读取完成后再继续
  Promise.all([readLogFile, readFixationsFile])
    .then(([log, fixations]) => {
      console.log(log, fixations)
      console.log(
        `使用的偏移值: timesetOffset=${timesetOffset}, leftOffset=${leftOffset}, topOffset=${topOffset}`
      )

      // 现在两个文件都已读取完成，可以安全地调用 createScene
      return createScene(babylonEngine, log, fixations)
    })
    .then(s => {
      scene = s

      // 设置渲染循环
      babylonEngine.runRenderLoop(() => {
        scene.render()
      })
    })
    .catch(error => {
      console.error('文件读取或场景创建出错:', error)
    })
}

function readCSVFromString (csvString, timesetOffset, leftOffset, topOffset) {
  const results = Papa.parse(csvString, {
    header: true,
    skipEmptyLines: true
  })

  const processedResults = results.data.map(data => {
    const timestamp = +data['start timestamp [ns]'].slice(0, 13) + timesetOffset
    const duration = +data['duration [ms]']
    const left = (+data['fixation x [px]'] - leftOffset) * 1.325
    const top = (+data['fixation y [px]'] - topOffset) * 1.325
    const index = data['fixation id']
    return { timestamp, duration, left, top, index }
  })

  return processedResults
}
