import {
  Scene,
  HemisphericLight,
  PointLight,
  UniversalCamera,
  Vector3,
  Color3,
  Color4,
  Quaternion,
  MeshBuilder,
  Ray,
  GlowLayer
} from '@babylonjs/core'
import * as gui from '@babylonjs/gui'
import calibrate from '../calibrate'
import createQuestionnaire from '../questionnaire'
import { createEnviroment } from '../utils'

//Create and export a function that takes a Babylon engine and returns a Babylon Scene
export async function createScene (engine, result, fixations) {
  //Create an empty Scene
  const scene = new Scene(engine)
  scene.clearColor = new Color4(0.078, 0.231, 0.427, 0.9)

  const pointLight = new PointLight('light', new Vector3(0, 3.4, 0), scene)
  pointLight.intensity = 0.8
  //Add some lighting
  let light = new HemisphericLight('light1', new Vector3(0, 10, 0), scene)
  light.diffuse = new Color3(1, 1, 1)
  light.specular = new Color3(1, 1, 1)
  light.groundColor = new Color3(1, 1, 1)

  //Add a camera and adjust its properties afterwards
  const camera = new UniversalCamera('camera', new Vector3(0, 0, 0), scene)
  camera.minZ = 0 // Adjust the distance of the camera's near plane
  // camera.position = new Vector3(-1.2, 0.2, -1)
  camera.fov = 1.45
  camera.attachControl(true) // Allow the camera to respond to user controls
  camera.inputs.attached.mouse.detachControl() // Disable mouse control
  let room, buttonHandlers

  calibrate(
    camera,
    scene,
    () => {
      room = createEnviroment(scene)
      buttonHandlers = createQuestionnaire(scene, {}, [], [], result.task)
    },
    true
  )

  var uiPlane = MeshBuilder.CreatePlane(
    'uiPlane',
    { width: 8, height: 4 },
    scene
  )
  uiPlane.position = new Vector3(0, -0.2, 2)

  uiPlane.renderingGroupId = 1

  uiPlane.parent = camera

  var advancedTexture = gui.AdvancedDynamicTexture.CreateForMesh(
    uiPlane,
    4096,
    2048,
    false
  )

  const glowLayer = new GlowLayer('glowLayer', scene)
  let currentlyHighlightedMesh = []
  glowLayer.intensity = 0.8 // Adjust intensity as needed
  glowLayer.blurKernelSize = 32 // Adjust blur for smoother effect

  // By default, glowLayer applies to all meshes, so let's set it to custom mode
  glowLayer.customEmissiveColorSelector = function (
    mesh,
    subMesh,
    material,
    result
  ) {
    // Only meshes explicitly added to our tracking will glow
    if (currentlyHighlightedMesh.includes(mesh)) {
      result.set(1, 0.6, 0, 1) // Orange glow with alpha 1
    } else {
      result.set(0, 0, 0, 0) // No glow
    }
    return true
  }

  const cameralLocations = result.cameralLocations.filter(
    d => d.timeStamp > result.calibrationTimeStamps[0]
  )

  const firstFrameTimestamp = cameralLocations[0].timeStamp
  const lastFrameTimestamp = cameralLocations.at(-1).timeStamp
  const usableFixations = fixations.filter(
    f =>
      f.timestamp < lastFrameTimestamp + 1000 &&
      f.timestamp > firstFrameTimestamp
  )

  let oldIndex = null

  function drawRedCircle (left, top, duration, index) {
    let circle = new gui.Ellipse()
    circle.width = '50px'
    circle.height = '50px'
    circle.color = 'red'
    circle.thickness = 3
    circle.background = 'transparent'
    circle.left = left
    circle.top = top
    advancedTexture.addControl(circle)

    let label = new gui.TextBlock()
    label.text = index.toString()
    label.color = 'red'
    label.fontSize = '25px'
    label.left = left + 30
    label.top = top + 30
    advancedTexture.addControl(label)

    setTimeout(function () {
      advancedTexture.removeControl(circle)
      advancedTexture.removeControl(label)
    }, duration)
  }

  function computeGazeRay (gazeX, gazeY, camera, uiPlane) {
    const normalizedX = gazeX / 1024 / 2 // Divide by 2 to get to -0.5 to 0.5 range
    const normalizedY = gazeY / 512 / 2 // Divide by 2 to get to -0.5 to 0.5 range

    const planeWidth = uiPlane.scaling.x * 4 // 4 is the width you specified when creating the plane
    const planeHeight = uiPlane.scaling.y * 2 // 2 is the height you specified when creating the plane

    // Calculate local position on the plane
    const localX = normalizedX * planeWidth
    const localY = -normalizedY * planeHeight // Negative because Y is typically inverted

    const localToWorldMatrix = uiPlane.getWorldMatrix()

    const localPos = new Vector3(localX, localY, 0)

    const worldPos = Vector3.TransformCoordinates(localPos, localToWorldMatrix)

    const eyePosition = camera.position.clone()
    const gazeDirection = worldPos.subtract(eyePosition).normalize()

    return {
      origin: eyePosition,
      direction: gazeDirection
    }
  }

  function analyzeUserGaze (gazeX, gazeY, camera, uiPlane, scene) {
    // Primary ray from gaze direction
    const gazeRay = computeGazeRay(gazeX, gazeY, camera, uiPlane)
    const origin = gazeRay.origin
    const direction = gazeRay.direction

    // Configuration for cone
    const coneAngle = 0.05 // radians (~3 degrees)
    const rayCount = 8 // Number of additional rays

    // Create perpendicular vectors to build cone
    const perp1 = Vector3.Cross(direction, Vector3.Up())
    if (perp1.lengthSquared() < 0.01) {
      // If direction is parallel to up vector, use a different reference
      perp1.copyFrom(Vector3.Cross(direction, Vector3.Right()))
    }
    perp1.normalize()

    const perp2 = Vector3.Cross(direction, perp1)
    perp2.normalize()

    // Store all hits
    let allPickInfos = []

    // Cast primary ray
    const primaryPickInfos = scene.multiPickWithRay(
      new Ray(origin, direction, 100),
      mesh => {
        return (
          mesh !== uiPlane &&
          mesh !== room &&
          mesh.name !== 'box' &&
          mesh.name !== 'sphere' &&
          mesh.name !== 'rootnode' &&
          mesh.name !== 'questionnairePlane'
        )
      }
    )

    if (primaryPickInfos && primaryPickInfos.length > 0) {
      allPickInfos = allPickInfos.concat(primaryPickInfos)
    }

    // Cast additional rays in a cone pattern
    for (let i = 0; i < rayCount; i++) {
      const angle = (i / rayCount) * Math.PI * 2

      // Calculate offset direction
      const offsetDir = new Vector3(
        direction.x +
          (Math.cos(angle) * perp1.x + Math.sin(angle) * perp2.x) * coneAngle,
        direction.y +
          (Math.cos(angle) * perp1.y + Math.sin(angle) * perp2.y) * coneAngle,
        direction.z +
          (Math.cos(angle) * perp1.z + Math.sin(angle) * perp2.z) * coneAngle
      ).normalize()

      const coneRay = new Ray(origin, offsetDir, 100)
      const conePickInfos = scene.multiPickWithRay(coneRay, mesh => {
        return (
          mesh !== uiPlane &&
          mesh !== room &&
          mesh.name !== 'box' &&
          mesh.name !== 'sphere' &&
          mesh.name !== 'rootnode' &&
          mesh.name !== 'questionnairePlane' &&
          mesh.name !== 'labelSPS' &&
          mesh.name !== 'lineSystem'
        )
      })

      if (conePickInfos && conePickInfos.length > 0) {
        allPickInfos = allPickInfos.concat(conePickInfos)
      }
    }

    // Reset currently highlighted meshes
    if (currentlyHighlightedMesh.length > 0) {
      currentlyHighlightedMesh = []

      glowLayer.customEmissiveColorSelector = function (
        mesh,
        subMesh,
        material,
        result
      ) {
        result.set(0, 0, 0, 0) // No glow for any mesh
        return true
      }
    }

    if (allPickInfos.length > 0) {
      // Remove duplicates (same mesh might be hit by multiple rays)
      const uniquePickInfos = []
      const seenMeshes = new Set()

      allPickInfos.forEach(pickInfo => {
        if (!seenMeshes.has(pickInfo.pickedMesh.id)) {
          seenMeshes.add(pickInfo.pickedMesh.id)
          uniquePickInfos.push(pickInfo)
        }
      })

      // Sort by distance (closest first)
      uniquePickInfos.sort((a, b) => a.distance - b.distance)

      console.log(
        `Found ${uniquePickInfos.length} objects intersecting with user's gaze`
      )

      // Process all intersections
      uniquePickInfos.forEach((pickInfo, index) => {
        currentlyHighlightedMesh.push(pickInfo.pickedMesh)

        console.log(`Intersection ${index + 1}:`)
        console.log(` - Mesh: ${pickInfo.pickedMesh.name}`)
        console.log(` - Distance: ${pickInfo.distance}`)
        console.log(
          ` - Point: (${pickInfo.pickedPoint.x.toFixed(
            2
          )}, ${pickInfo.pickedPoint.y.toFixed(
            2
          )}, ${pickInfo.pickedPoint.z.toFixed(2)})`
        )
      })

      // Update glow layer for all highlighted meshes
      glowLayer.customEmissiveColorSelector = function (
        mesh,
        subMesh,
        material,
        result
      ) {
        if (currentlyHighlightedMesh.includes(mesh)) {
          result.set(1, 0.6, 0, 1) // Orange glow
        } else {
          result.set(0, 0, 0, 0) // No glow
        }
        return true
      }

      // Return the results
      return {
        gazedObjects: uniquePickInfos,
        primaryTarget: uniquePickInfos[0].pickedMesh
      }
    } else {
      console.log("User's gaze doesn't intersect with any objects")
      return {
        gazedObjects: [],
        primaryTarget: null
      }
    }
  }

  scene.onReadyObservable.add(() => {
    for (let i = 0; i < cameralLocations.length; i++) {
      setTimeout(() => {
        // Restore the camera transform
        camera.position = Vector3.FromArray(cameralLocations[i].position)
        camera.rotationQuaternion = Quaternion.FromArray(
          cameralLocations[i].rotation
        )

        const buttonClickingHistory = cameralLocations[i].buttonClickingHistory
        const { index } = buttonClickingHistory

        if (index !== oldIndex) {
          oldIndex = index
          updateQuestionnaire(buttonClickingHistory, buttonHandlers)
        }

        // Tools.CreateScreenshot(engine, camera, { precision: 1.0 })
      }, cameralLocations[i].timeStamp - firstFrameTimestamp)
    }

    setTimeout(() => {
      buttonHandlers.next && buttonHandlers.next(false)
    }, lastFrameTimestamp - firstFrameTimestamp + 20)

    for (let i = 0; i < usableFixations.length; i++) {
      const { left, top, duration, timestamp, index } = usableFixations[i]

      setTimeout(() => {
        drawRedCircle(left, top, duration, index)
        analyzeUserGaze(left, top, camera, uiPlane, scene)
      }, timestamp - firstFrameTimestamp)
    }
  })

  return scene
}

function updateQuestionnaire (buttonClickingHistory, buttonHandlers) {
  const { key, index } = buttonClickingHistory
  index > 0 && buttonHandlers[key] && buttonHandlers[key]()
}
