/**
 * Generates three graph analysis questions based on a graph object
 * @param {Object} graph - Graph object containing nodes and links arrays
 * @returns {Array} Array of 3 question objects
 */
function generateGraphQuestions (graph) {
  // Create adjacency list representation
  const adjacencyList = buildAdjacencyList(graph)

  // Generate question 1: Shortest path between two nodes
  const shortestPathQuestion = generateShortestPathQuestion(
    graph,
    adjacencyList
  )

  // Generate question 2: Highest connectivity
  const highestConnectivityQuestion = generateHighestConnectivityQuestion(
    graph,
    adjacencyList
  )

  // Generate question 3: Common neighbors
  const commonNeighborsQuestion = generateCommonNeighborsQuestion(
    graph,
    adjacencyList
  )

  return [
    shortestPathQuestion,
    highestConnectivityQuestion,
    commonNeighborsQuestion
  ]
}

/**
 * Builds an adjacency list from the graph
 * @param {Object} graph - Graph object containing nodes and links arrays
 * @returns {Object} Adjacency list where keys are node indices and values are arrays of neighboring node indices
 */
function buildAdjacencyList (graph) {
  const adjacencyList = {}

  // Initialize empty arrays for all nodes
  graph.nodes.forEach(node => {
    adjacencyList[node.index] = new Set()
  })

  // Add connections based on links
  graph.links.forEach(link => {
    const sourceIndex =
      typeof link.source === 'object' ? link.source.index : link.source
    const targetIndex =
      typeof link.target === 'object' ? link.target.index : link.target

    if (sourceIndex === targetIndex) {
      return
    }

    adjacencyList[sourceIndex].add(targetIndex)
    adjacencyList[targetIndex].add(sourceIndex) // Assuming undirected graph
  })

  graph.nodes.forEach(node => {
    adjacencyList[node.index] = Array.from(adjacencyList[node.index])
  })

  return adjacencyList
}

/**
 * Calculates shortest path length between two nodes using BFS
 * @param {Object} adjacencyList - Adjacency list representation of the graph
 * @param {number} startNode - Starting node index
 * @param {number} endNode - Target node index
 * @returns {number} Length of shortest path, or -1 if no path exists
 */
function shortestPathLength (adjacencyList, startNode, endNode) {
  if (startNode === endNode) return 0

  const visited = new Set()
  const queue = [[startNode, 0]] // [node, distance]
  visited.add(startNode)

  while (queue.length > 0) {
    const [node, distance] = queue.shift()

    for (const neighbor of adjacencyList[node]) {
      if (neighbor === endNode) {
        return distance + 1
      }

      if (!visited.has(neighbor)) {
        visited.add(neighbor)
        queue.push([neighbor, distance + 1])
      }
    }
  }

  return -1 // No path found
}

/**
 * Finds node with highest degree (most connections)
 * @param {Object} adjacencyList - Adjacency list representation of the graph
 * @returns {Object} Object containing the node index with highest degree and its degree
 */
function findHighestDegreeNode (adjacencyList) {
  let maxDegree = -1
  let maxDegreeNode = null

  for (const [node, neighbors] of Object.entries(adjacencyList)) {
    const degree = neighbors.length
    if (degree > maxDegree) {
      maxDegree = degree
      maxDegreeNode = parseInt(node)
    }

    if (degree >= maxDegree) {
      console.log(node, degree, neighbors)
    }
  }

  return { node: maxDegreeNode, degree: maxDegree }
}

/**
 * Counts common neighbors between two nodes
 * @param {Object} adjacencyList - Adjacency list representation of the graph
 * @param {number} node1 - First node index
 * @param {number} node2 - Second node index
 * @returns {number} Count of common neighbors
 */
function countCommonNeighbors (adjacencyList, node1, node2) {
  const neighbors1 = new Set(adjacencyList[node1])
  const neighbors2 = new Set(adjacencyList[node2])

  // Remove the nodes themselves from each other's neighbor sets
  neighbors1.delete(node2)
  neighbors2.delete(node1)

  // Count common elements
  let commonCount = 0
  for (const neighbor of neighbors1) {
    if (neighbors2.has(neighbor)) {
      commonCount++
    }
  }

  return commonCount
}

/**
 * Finds pairs of nodes with shortest path length between 1-9
 * @param {Object} graph - Graph object containing nodes and links
 * @param {Object} adjacencyList - Adjacency list representation of the graph
 * @returns {Array} Array of node pairs with valid shortest path lengths
 */
function findValidPathPairs (graph, adjacencyList) {
  const validPairs = []
  const nodeIndices = graph.nodes.map(node => node.index)

  for (let i = 0; i < nodeIndices.length; i++) {
    for (let j = i + 1; j < nodeIndices.length; j++) {
      const node1 = nodeIndices[i]
      const node2 = nodeIndices[j]

      const pathLength = shortestPathLength(adjacencyList, node1, node2)
      if (pathLength >= 4 && pathLength <= 5) {
        validPairs.push({ nodes: [node1, node2], pathLength })
      }
    }
  }

  console.log(validPairs)

  return validPairs
}

/**
 * Generates shortest path question
 * @param {Object} graph - Graph object containing nodes and links
 * @param {Object} adjacencyList - Adjacency list representation of the graph
 * @returns {Object} Question object
 */
function generateShortestPathQuestion (graph, adjacencyList) {
  const validPairs = findValidPathPairs(graph, adjacencyList)

  // If no valid pairs found, create a fallback pair
  if (validPairs.length === 0) {
    console.warn(
      'No node pairs with shortest path between 1-9 found. Using first two nodes as fallback.'
    )
    const highlightedNodes = [graph.nodes[0].index, graph.nodes[1].index]
    const pathLength = shortestPathLength(
      adjacencyList,
      highlightedNodes[0],
      highlightedNodes[1]
    )

    return {
      text: 'How long is the shortest path between the two highlighted nodes?',
      answers: ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9'],
      highlightedNodes,
      answer: pathLength,
      graph
    }
  }

  // Select a random valid pair
  const randomPair = validPairs[Math.floor(Math.random() * validPairs.length)]

  return {
    text: 'How long is the shortest path between the two highlighted nodes?',
    answers: ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9'],
    highlightedNodes: randomPair.nodes,
    answer: randomPair.pathLength,
    graph
  }
}

/**
 * Generates highest connectivity question
 * @param {Object} adjacencyList - Adjacency list representation of the graph
 * @returns {Object} Question object
 */
function generateHighestConnectivityQuestion (graph, adjacencyList) {
  const { node: highestDegreeNode } = findHighestDegreeNode(adjacencyList)

  return {
    text: 'Which node has the highest number of edges to other nodes?',
    answers: ['Enter node label'],
    highlightedNodes: [],
    answer: highestDegreeNode,
    graph
  }
}

/**
 * Finds pairs of nodes with common neighbors count between 0-9
 * @param {Object} adjacencyList - Adjacency list representation of the graph
 * @param {Array} nodeIndices - Array of node indices
 * @returns {Array} Array of node pairs with valid common neighbor counts
 */
function findValidCommonNeighborPairs (adjacencyList, nodeIndices) {
  const validPairs = []

  for (let i = 0; i < nodeIndices.length; i++) {
    for (let j = i + 1; j < nodeIndices.length; j++) {
      const node1 = nodeIndices[i]
      const node2 = nodeIndices[j]

      const commonCount = countCommonNeighbors(adjacencyList, node1, node2)
      if (commonCount >= 2 && commonCount <= 8) {
        validPairs.push({ nodes: [node1, node2], commonCount })
      }
    }
  }

  console.log(validPairs)

  return validPairs
}

/**
 * Generates common neighbors question
 * @param {Object} graph - Graph object containing nodes and links
 * @param {Object} adjacencyList - Adjacency list representation of the graph
 * @returns {Object} Question object
 */
function generateCommonNeighborsQuestion (graph, adjacencyList) {
  const nodeIndices = graph.nodes.map(node => node.index)
  const validPairs = findValidCommonNeighborPairs(adjacencyList, nodeIndices)

  // If no valid pairs found, create a fallback pair
  if (validPairs.length === 0) {
    console.warn(
      'No node pairs with common neighbors between 0-9 found. Using first two nodes as fallback.'
    )
    const highlightedNodes = [nodeIndices[0], nodeIndices[1]]
    const commonCount = countCommonNeighbors(
      adjacencyList,
      highlightedNodes[0],
      highlightedNodes[1]
    )

    return {
      text: 'How many neighbors do the two highlighted nodes have in common?',
      answers: ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9'],
      highlightedNodes,
      answer: commonCount,
      graph
    }
  }

  // Select a random valid pair
  const randomPair = validPairs.sort((a, b) => b.commonCount - a.commonCount)[0]
  console.log(randomPair)

  return {
    text: 'How many neighbors do the two highlighted nodes have in common?',
    answers: ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9'],
    highlightedNodes: randomPair.nodes,
    answer: randomPair.commonCount,
    graph
  }
}

export default generateGraphQuestions
